# Copyright (c) 2016-2017 Anki, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License in the file LICENSE.txt or at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''Engine connection.

The SDK operates by connecting to the Cozmo "engine" - typically the Cozmo
app that runs on an iOS or Android device.

The engine is responsible for much of the work that Cozmo does, including
image recognition, path planning, behaviors and animation handling, etc.

The :mod:`cozmo.run` module takes care of opening a connection over a USB
connection to a device, but the :class:`CozmoConnection` class defined in
this module does the work of relaying messages to and from the engine and
dispatching them to the :class:`cozmo.robot.Robot` instance.
'''

# __all__ should order by constants, event classes, other classes, functions.
__all__ = ['EvtRobotFound', 'CozmoConnection']


import asyncio
import platform

import cozmoclad

from . import logger
from . import anim
from . import clad_protocol
from . import event
from . import exceptions
from . import robot
from . import version

from . import _clad
from ._clad import _clad_to_engine_cozmo, _clad_to_engine_iface, _clad_to_game_cozmo, _clad_to_game_iface


class EvtConnected(event.Event):
    '''Triggered when the initial connection to the device has been established.

    This connection is setup before contacting the robot - Wait for EvtRobotFound
    or EvtRobotReady for a usefully configured Cozmo instance.
    '''
    conn = 'The connected CozmoConnection object'


class EvtRobotFound(event.Event):
    '''Triggered when a Cozmo robot is detected, but before he's initialized.

    :class:`cozmo.robot.EvtRobotReady` is dispatched when the robot is fully initialized.
    '''
    robot = 'The Cozmo object for the robot'


class EvtConnectionClosed(event.Event):
    '''Triggered when the connection to the controlling device is closed.
    '''
    exc = 'The exception that triggered the closure, or None'


# Some messages have no robotID but should still be forwarded to the primary robot
FORCED_ROBOT_MESSAGES = {"AnimationAborted",
                         "AnimationEvent",
                         "BehaviorObjectiveAchieved",
                         "BehaviorTransition",
                         "BlockPickedUp",
                         "BlockPlaced",
                         "BlockPoolDataMessage",
                         "CarryStateUpdate",
                         "ChargerEvent",
                         "CubeLightsStateTransition",
                         "CurrentCameraParams",
                         "LoadedKnownFace",
                         "ObjectPowerLevel",
                         "ObjectProjectsIntoFOV",
                         "ObjectStates",
                         "ReactionaryBehaviorTransition",
                         "RobotChangedObservedFaceID",
                         "RobotCliffEventFinished",
                         "RobotErasedAllEnrolledFaces",
                         "RobotErasedEnrolledFace",
                         "RobotObservedMotion",
                         "RobotObservedPet",
                         "RobotObservedPossibleObject",
                         "RobotOnChargerPlatformEvent",
                         "RobotReachedEnrollmentCount",
                         "RobotRenamedEnrolledFace",
                         "UnexpectedMovement"}


class CozmoConnection(event.Dispatcher, clad_protocol.CLADProtocol):
    '''Manages the connection to the Cozmo app to communicate with the core engine.

    An instance of this class is passed to functions used with
    :func:`cozmo.run.connect`.  At the point the function is executed,
    the connection is already established and verified, and the
    :class:`EvtConnected` has already been sent.

    However, after the initial connection is established, programs will usually
    want to call :meth:`wait_for_robot` to wait for an actual Cozmo robot to
    be detected and initialized before doing useful work.
    '''

    #: callable: The factory function that returns a
    #: :class:`cozmo.robot.Robot` class or subclass instance.
    robot_factory = robot.Robot

    #: callable: The factory function that returns an
    #: :class:`cozmo.anim.AnimationNames` class or subclass instance.
    anim_names_factory = anim.AnimationNames

    # overrides for CLADProtocol
    clad_decode_union = _clad_to_game_iface.MessageEngineToGame
    clad_encode_union = _clad_to_engine_iface.MessageGameToEngine

    def __init__(self, *a, **kw):
        super().__init__(*a, **kw)
        self._is_connected = False
        self._is_ui_connected = False
        self._running = True
        self._robots = {}
        self._primary_robot = None

        #: A dict containing information about the device the connection is using.
        self.device_info = {}

        #: An :class:`cozmo.anim.AnimationNames` object that references all
        #: available animation names
        self.anim_names = self.anim_names_factory(self)


    #### Private Methods ####

    def __repr__(self):
        info = ' '.join(['%s="%s"' % (k, self.device_info[k])
            for k in sorted(self.device_info.keys())])
        return '<%s %s>' % (self.__class__.__name__, info)

    def connection_made(self, transport):
        super().connection_made(transport)
        self._is_connected = True

    def connection_lost(self, exc):
        super().connection_lost(exc)
        self._is_connected = False
        if self._running:
            self.abort(exceptions.ConnectionAborted("Lost connection to the device"))
            logger.error("Lost connection to the device: %s", exc)

    async def shutdown(self):
        '''Close the connection to the device.'''
        if self._running and self._is_connected:
            logger.info("Shutting down connection")
            self._running = False
            event._abort_futures(exceptions.SDKShutdown())
            self._stop_dispatcher()
            self.transport.close()

    def abort(self, exc):
        '''Abort the connection to the device.'''
        if self._running:
            logger.info('Aborting connection: %s', exc)
            self._running = False
            # Allow any currently pending futures to complete before the
            # remainder are aborted.
            self._loop.call_soon(lambda: event._abort_futures(exc))
            self._stop_dispatcher()
            self.transport.close()


    def msg_received(self, msg):
        '''Receives low level communication messages from the engine.'''
        if not self._running:
            return

        try:
            tag_name = msg.tag_name

            if tag_name == 'Ping':
                # short circuit to avoid unnecessary event overhead
                return self._handle_ping(msg._data)

            elif tag_name == 'UiDeviceConnected':
                # handle outside of event dispatch for quick abort in case
                # of a version mismatch problem.
                return self._handle_ui_device_connected(msg._data)

            msg = msg._data
            robot_id = getattr(msg, 'robotID', None)

            if not robot_id and self._primary_robot and (tag_name in FORCED_ROBOT_MESSAGES):
                # Forward to the primary robot
                robot_id = self._primary_robot.robot_id

            event_name = '_Msg' +  tag_name

            evttype = getattr(_clad, event_name, None)
            if evttype is None:
                logger.error('Received unknown CLAD message %s', event_name)
                return

            if robot_id:
                # dispatch robot-specific messages to Cozmo robot instances
                return self._process_robot_msg(robot_id, evttype, msg)

            self.dispatch_event(evttype, msg=msg)

        except Exception as exc:
            # No exceptions should reach this point; it's a bug if they do.
            self.abort(exc)

    def _process_robot_msg(self, robot_id, evttype, msg):
        if robot_id > 1:
            # One day we might support multiple robots.. if we see a robot_id != 1
            # currently though, it's an error.
            # Note: MsgRobotPoked always sends the wrong id through currently
            logger.debug('INVALID ROBOT_ID SEEN robot_id=%s event=%s msg=%s', robot_id, evttype, msg.__str__())
            robot_id = 1 # XXX remove when errant messages have been fixed

        robot = self._robots.get(robot_id)
        if not robot:
            logger.info('Found robot id=%s', robot_id)
            robot = self.robot_factory(self, robot_id, is_primary=self._primary_robot is None)
            self._robots[robot_id] = robot
            if not self._primary_robot:
                self._primary_robot = robot
            # Dispatch an event notifying that a new robot has been found
            # the robot itself will send EvtRobotReady after initialization
            self.dispatch_event(EvtRobotFound, robot=robot)

            # _initialize will set the robot to a known good state in the
            # background and dispatch a EvtRobotReady event when completed.
            robot._initialize()

        robot.dispatch_event(evttype, msg=msg)


    #### Properties ####

    @property
    def is_connected(self):
        '''bool: True if currently connected to the remote engine.'''
        return self._is_connected


    #### Private Event handlers ####

    def _handle_ping(self, msg):
        '''Respond to a ping event.'''
        if msg.isResponse:
            # To avoid duplication, pings originate from engine, and engine
            # accumulates the latency info from the responses
            logger.error("Only engine should receive responses")
        else:
            resp = _clad_to_engine_iface.Ping(
                counter=msg.counter,
                timeSent_ms=msg.timeSent_ms,
                isResponse=True)
            self.send_msg(resp)

    def _recv_default_handler(self, event, **kw):
        '''Default event handler.'''
        if event.event_name.startswith('msg_animation'):
            return self.anim.dispatch_event(event)

        logger.debug('Engine received unhandled event_name=%s  kw=%s', event, kw)

    def _recv_msg_animation_available(self, evt, msg):
        self.anim_names.dispatch_event(evt)

    def _recv_msg_end_of_message(self, evt, *a, **kw):
        self.anim_names.dispatch_event(evt)

    def _handle_ui_device_connected(self, msg):
        if msg.connectionType != _clad_to_engine_cozmo.UiConnectionType.SdkOverTcp:
            # This isn't for us
            return

        if msg.deviceID != 1:
            logger.error('Unexpected Device Id %s', msg.deviceID)
            return

        # Verify that engine and SDK are compatible
        clad_hashes_match = False
        try:
            cozmoclad.assert_clad_match(msg.toGameCLADHash, msg.toEngineCLADHash)
            clad_hashes_match = True
        except cozmoclad.CLADHashMismatch as exc:
            logger.error(exc)

        build_versions_match = (cozmoclad.__build_version__ == '00000.00000.00000'
            or cozmoclad.__build_version__ == msg.buildVersion)

        if clad_hashes_match and not build_versions_match:
            # If CLAD hashes match, and this is only a minor version change,
            # then still allow connection (it's just an app hotfix
            # that didn't require CLAD or SDK changes)
            sdk_major_version = cozmoclad.__build_version__.split(".")[0:2]
            build_major_version = msg.buildVersion.split(".")[0:2]
            build_versions_match = (sdk_major_version == build_major_version)

        if clad_hashes_match and build_versions_match:
            connection_success_msg = _clad_to_engine_iface.UiDeviceConnectionSuccess(
                connectionType=msg.connectionType,
                deviceID=msg.deviceID,
                buildVersion = cozmoclad.__version__,
                sdkModuleVersion = version.__version__,
                pythonVersion = platform.python_version(),
                pythonImplementation = platform.python_implementation(),
                osVersion = platform.platform(),
                cpuVersion = platform.machine())

            self.send_msg(connection_success_msg)

        else:
            try:
                wrong_version_msg = _clad_to_engine_iface.UiDeviceConnectionWrongVersion(
                    reserved=0,
                    connectionType=msg.connectionType,
                    deviceID = msg.deviceID,
                    buildVersion = cozmoclad.__version__)

                self.send_msg(wrong_version_msg)
            except AttributeError:
                pass

            line_separator = "=" * 80
            error_message = "\n" + line_separator + "\n"

            if not build_versions_match:

                def _trimmed_version(ver_string):
                    # Trim leading zeros from the version string.
                    trimmed_string = ""
                    for i in ver_string.split("."):
                        trimmed_string += str(int(i)) + "."
                    return trimmed_string[:-1]  # remove trailing "."

                error_message += ("App and SDK versions do not match!\n"
                                  "----------------------------------\n"
                                  "SDK's cozmoclad version: %s\n"
                                  "         != app version: %s\n\n"
                                  % (cozmoclad.__version__, _trimmed_version(msg.buildVersion)))

                if cozmoclad.__build_version__ < msg.buildVersion:
                    # App is newer
                    error_message += ('Please update your SDK to the newest version by calling command:\n'
                                      '"pip3 install --user --upgrade cozmo"\n'
                                      'and downloading the latest examples from:\n'
                                      'http://cozmosdk.anki.com/docs/downloads.html\n')
                else:
                    # SDK is newer
                    error_message += ('Please either:\n\n'
                                      '1) Update your app to the most recent version on the app store.\n'
                                      '2) Or, if you prefer, please determine which SDK version matches\n'
                                      '   your app version at: http://go.anki.com/cozmo-sdk-version\n'
                                      '   Then downgrade your SDK by calling the following command,\n'
                                      '   replacing SDK_VERSION with the version listed at that page:\n'
                                      '   "pip3 install --ignore-installed cozmo==SDK_VERSION"\n')

            else:
                # CLAD version mismatch
                error_message += ('CLAD Hashes do not match!\n'
                                  '-------------------------\n'
                                  'Your Python and C++ CLAD versions do not match - connection refused.\n'
                                  'Please check that you have the most recent versions of both the SDK and the\n'
                                  'Cozmo app. You may update your SDK by calling:\n'
                                  '"pip3 install --user --ignore-installed cozmo".\n'
                                  'Please also check the app store for a Cozmo app update.\n')

            error_message += line_separator
            logger.error(error_message)

            exc = exceptions.SDKVersionMismatch("SDK library does not match software running on device")
            self.abort(exc)
            return

        self._is_ui_connected = True
        self.dispatch_event(EvtConnected, conn=self)

        logger.info('App connection established. sdk_version=%s '
                'cozmoclad_version=%s app_build_version=%s',
                version.__version__, cozmoclad.__version__, msg.buildVersion)

        # We send RequestObjectStates before refreshing the animation names
        # as this ensures that we will receive the responses before we mark the
        # robot as ready
        msg = _clad_to_engine_iface.RequestObjectStates()
        self.send_msg(msg)

        self.anim_names.refresh()

    def _recv_msg_image_chunk(self, evt, *, msg):
        if self._primary_robot:
            self._primary_robot.dispatch_event(evt)


    #### Public Event Handlers ####

    #### Commands ####

    async def _wait_for_robot(self, timeout=5):
        if not self._primary_robot:
            await self.wait_for(EvtRobotFound, timeout=timeout)
        if self._primary_robot.is_ready:
            return self._primary_robot
        await self._primary_robot.wait_for(robot.EvtRobotReady, timeout=timeout)
        return self._primary_robot

    async def wait_for_robot(self, timeout=5):
        '''Wait for a Cozmo robot to connect and complete initialization.

        Args:
            timeout (float): Maximum length of time to wait for a robot to be ready in seconds.
        Returns:
            A :class:`cozmo.robot.Robot` instance that's ready to use.
        Raises:
            :class:`asyncio.TimeoutError` if there's no response from the robot.
        '''
        try:
            robot = await self._wait_for_robot(timeout)
            if robot and robot.drive_off_charger_on_connect:
                await robot.drive_off_charger_contacts().wait_for_completed()
        except asyncio.TimeoutError:
            logger.error('Timed out waiting for robot to initialize')
            raise
        return robot



# Copyright (c) 2016 Anki, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License in the file LICENSE.txt or at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''Event dispatch system.

The SDK is based around the dispatch and observation of events.
Objects inheriting from the :class:`Dispatcher` generate and
dispatch events as the state of the robot and its world are updated.

For example the :class:`cozmo.objects.LightCube` class generates an
:class:`~cozmo.objects.EvtObjectTapped` event anytime the cube the object
represents is tapped.

The event can be observed in a number of different ways:

#. By calling the :meth:`~Dispatcher.wait_for` method on the object to observe.
   This will wait until the specific event has been sent to that object and
   return the generated event.
#. By calling :meth:`~Dispatcher.add_event_handler` on the object
   to observe, which will cause the supplied function to be called every time
   the specified event occurs (use the :func:`oneshot` decorator
   to only have the handler called once)
#. By sub-classing a type and implementing a receiver method.
   For example, subclass the :class:`cozmo.objects.LightCube` type and implement `evt_object_tapped`.
   Note that the factory attribute would need to be updated on the
   generating class for your type to be used by the SDK.
   For example, :attr:`~cozmo.world.World.light_cube_factory` in this example.
#. By subclassing a type and implementing a default receiver method.
   Events not dispatched to an explicit receiver method are dispatched to
   `recv_default_handler`.

Events are dispatched to a target object (by calling :meth:`dispatch_event`
on the receiving object).  In line with the above, upon receiving an event,
the object will:

#. Dispatch the event to any handlers which have explicitly registered interest
   in the event (or a superclass of the event) via
   :meth:`~Dispatcher.add_event_handler` or via :meth:`Dispatcher.wait_for`
#. Dispatch the event to any "children" of the object (see below)
#. Dispatch the event to method handlers on the receiving object, or the
   `recv_default_handler` if it has no matching handler
#. Dispatch the event to the parent of the object (if any), and in turn onto
   the parent's parents.

Any handler may raise a :class:`~cozmo.exceptions.StopPropogation` exception
to prevent the event reaching any subsequent handlers (but generally should
have no need to do so).

Child objects receive all events that are sent to the originating object
(which may have multiple children).

Originating objects may have one parent object, which receives all events sent
to its child.

For example, :class:`cozmo.robot.Robot` creates a :class:`cozmo.world.World`
object and sets itself as a parent and the World as the child; both receive
events sent to the other.

The World class creates individual :class:`cozmo.objects.ObservableObject` objects
as they are discovered and makes itself a parent, so as to receive all events
sent to the child.  However, it does not make those ObservableObject objects children
for the sake of message dispatch as they only need to receive a small subset
of messages the World object receives.
'''

# __all__ should order by constants, event classes, other classes, functions.
__all__ = ['Event', 'Dispatcher', 'Filter', 'Handler',
    'oneshot', 'filter_handler', 'wait_for_first']


import asyncio
import collections
import inspect
import re
import weakref

from . import base
from . import exceptions
from . import logger


# from https://stackoverflow.com/questions/1175208/elegant-python-function-to-convert-camelcase-to-snake-case
_first_cap_re = re.compile('(.)([A-Z][a-z]+)')
_all_cap_re = re.compile('([a-z0-9])([A-Z])')
def _uncamelcase(name):
    s1 = _first_cap_re.sub(r'\1_\2', name)
    return _all_cap_re.sub(r'\1_\2', s1).lower()

registered_events = {}

active_dispatchers = weakref.WeakSet()

class _rprop:
    def __init__(self, value):
        self._value = value
    def __get__(self, instance, owner):
        return self._value

class docstr(str):
    @property
    def __doc__(self):
        return self.__str__()


class _AutoRegister(type):
    '''helper to automatically register event classes wherever they're defined
    without requiring a class decorator'''

    def __new__(mcs, name, bases, attrs, **kw):
        if name in ('Event',):
            return super().__new__(mcs, name, bases, attrs, **kw)

        if not (name.startswith('Evt') or name.startswith('_Evt') or name.startswith('_Msg')):
            raise ValueError('Event class names must begin with "Evt (%s)"' % name)

        if '__doc__' not in attrs:
            raise ValueError('Event classes must have a docstring')

        props = set()
        for base in bases:
            if hasattr(base, '_props'):
                props.update(base._props)

        newattrs = {'_internal': False}
        for k, v in attrs.items():
            if k[0] == '_':
                newattrs[k] = v
                continue
            if k in props:
                raise ValueError("Event class %s duplicates property %s defined in superclass" % (self, k))
            props.add(k)
            newattrs[k] = docstr(v)
        newattrs['_props'] = props
        newattrs['_props_sorted'] = sorted(props)

        if name[0] == '_':
            newattrs['_internal'] = True
            name = name[1:]

        # create a read only property for the event name
        newattrs['event_name'] = _rprop(name)
        return super().__new__(mcs, name, bases, newattrs, **kw)

    def __init__(cls, name, bases, attrs, **kw):
        if name in registered_events:
            raise ValueError("Duplicate event name %s (%s duplicated by %s)"
                    % (name, _full_qual_name(cls), _full_qual_name(registered_events[name])))
        registered_events[name] = cls
        return super().__init__(name, bases, attrs, **kw)


def _full_qual_name(obj):
    return obj.__module__ + '.' + obj.__qualname__


class Event(metaclass=_AutoRegister):
    '''An event representing an action that has occurred.

    Instances of an Event have attributes set to values passed to the event.

    For example, :class:`cozmo.objects.EvtObjectTapped` defines obj and tap_count
    parameters which can be accessed as ``evt.obj`` and ``evt.tap_count``.
    '''

    #_first_raised_by = "The object that generated the event"
    #_last_raised_by = "The object that last relayed the event to the dispatched handler"

    def __init__(self, **kwargs):
        unset = self._props.copy()
        for k, v in kwargs.items():
            if k not in self._props:
                raise ValueError("Event %s has no parameter called %s" % (self.event_name, k))
            setattr(self, k, v)
            unset.remove(k)
        for k in unset:
            setattr(self, k, None)
        self._delivered_to = set()

    def __repr__(self):
        kvs = {'name': self.event_name}
        for k in self._props_sorted:
            kvs[k] = getattr(self, k)
        return '<%s %s>' % (self.__class__.__name__, ' '.join(['%s=%s' % kv for kv in kvs.items()]),)

    def _params(self):
        return {k: getattr(self, k) for k in self._props}

    @classmethod
    def _handler_method_name(cls):
        name = 'recv_' + _uncamelcase(cls.event_name)
        if cls._internal:
            name = '_' + name
        return name

    def _dispatch_to_func(self, f):
        return f(self, **self._params())

    def _dispatch_to_obj(self, obj, fallback_to_default=True):
        for cls in self._parent_event_classes():
            f = getattr(obj, cls._handler_method_name(), None)
            if f and not self._is_filtered(f):
                return self._dispatch_to_func(f)

        if fallback_to_default:
            name = 'recv_default_handler'
            if self._internal:
                name = '_' + name
            f = getattr(obj, name, None)
            if f and not self._is_filtered(f):
                return f(self, **self._params())

    def _dispatch_to_future(self, fut):
        if not fut.done():
            fut.set_result(self)

    def _is_filtered(self, f):
        filters = getattr(f, '_handler_filters', None)
        if filters is None:
            return False
        for filter in filters:
            if filter(self):
                return False
        return True

    def _parent_event_classes(self):
        for cls in self.__class__.__mro__:
            if cls != Event and issubclass(cls, Event):
                yield cls


def _register_dynamic_event_type(event_name, attrs):
    return type(event_name, (Event,), attrs)


class Handler(collections.namedtuple('Handler', 'obj evt f')):
    '''A Handler is returned by :meth:`Dispatcher.add_event_handler`

    The handler can be disabled at any time by calling its :meth:`disable`
    method.
    '''
    __slots__ = ()

    def disable(self):
        '''Removes the handler from the object it was originally registered with.'''
        return self.obj.remove_event_handler(self.evt, self.f)

    @property
    def oneshot(self):
        '''bool: True if the wrapped handler function will only be called once.'''
        return getattr(self.f, '_oneshot_handler', False)


class NullHandler(Handler):
    def disable(self):
        pass


class Dispatcher(base.Base):
    '''Mixin to provide event dispatch handling.'''

    def __init__(self, *a, dispatch_parent=None, loop=None, **kw):
        super().__init__(**kw)
        active_dispatchers.add(self)
        self._dispatch_parent = dispatch_parent
        self._dispatch_children = []
        self._dispatch_handlers = collections.defaultdict(list)
        if not loop:
            raise ValueError("Loop was not supplied to "+self.__class__.__name__)
        self._loop = loop or asyncio.get_event_loop()
        self._dispatcher_running = True

    def _set_parent_dispatcher(self, parent):
        self._dispatch_parent = parent

    def _add_child_dispatcher(self, child):
        self._dispatch_children.append(child)

    def _stop_dispatcher(self):
        """Stop dispatching events - call before closing the connection to prevent stray dispatched events"""
        self._dispatcher_running = False

    def add_event_handler(self, event, f):
        """Register an event handler to be notified when this object receives a type of Event.

        Expects a subclass of Event as the first argument.  If the class has
        subclasses then the handler will be notified for events of that subclass too.
        For example, adding a handler for :class:`~cozmo.action.EvtActionCompleted` 
        will cause the handler to also be notified for
        :class:`~cozmo.anim.EvtAnimationCompleted` as it's a subclass.

        Callable handlers (e.g. functions) are called with a first argument
        containing an Event instance and the remaining keyword arguments set as
        the event parameters.

        For example, ``def my_ontap_handler(evt, *, obj, tap_count, **kwargs)``
        or ``def my_ontap_handler(evt, obj=None, tap_count=None, **kwargs)``

        It's recommended that a ``**kwargs`` parameter be included in the
        definition so that future expansion of event parameters do not cause
        the handler to fail.

        Callable handlers may raise an events.StopProgation exception to prevent
        other handlers listening to the same event from being triggered.

        :class:`asyncio.Future` handlers are called with a result set to the event.

        Args:
            event (:class:`Event`): A subclass of :class:`Event` (not an instance of that class)
            f (callable): A callable or :class:`asyncio.Future` to execute when the event is received
        Raises:
            :class:`TypeError`: An invalid event type was supplied

        """
        if not issubclass(event, Event):
            raise TypeError("event must be a subclass of Event (not an instance)")

        if not self._dispatcher_running:
            return NullHandler(self, event, f)

        if isinstance(f, asyncio.Future):
            # futures can only be called once.
            f = oneshot(f)

        handler = Handler(self, event, f)
        self._dispatch_handlers[event.event_name].append(handler)
        return handler

    def remove_event_handler(self, event, f):
        """Remove an event handler for this object.

        Args:
            event (:class:`Event`): The event class, or an instance thereof,
                used with register_event_handler.
            f (callable or :class:`Handler`): The callable object that was
                passed as a handler to :meth:`add_event_handler`, or a
                :class:`Handler` instance that was returned by
                :meth:`add_event_handler`.
        Raises:
            :class:`ValueError`: No matching handler found.
        """
        if not (isinstance(event, Event) or (isinstance(event, type) and  issubclass(event, Event))):
            raise TypeError("event must be a subclasss or instance of Event")

        if isinstance(f, Handler):
            for i, h in enumerate(self._dispatch_handlers[event.event_name]):
                if h == f:
                    del self._dispatch_handlers[event.event_name][i]
                    return
        else:
            for i, h in enumerate(self._dispatch_handlers[event.event_name]):
                if h.f == f:
                    del self._dispatch_handlers[event.event_name][i]
                    return
        raise ValueError("No matching handler found for %s (%s)" % (event.event_name, f) )

    def dispatch_event(self, event, **kw):
        '''Dispatches a single event to registered handlers.

        Not generally called from user-facing code.

        Args:
            event (:class:`Event`): An class or instance of :class:`Event`
            kw (dict): If a class is passed to event, then the remaining keywords
                are passed to it to create an instance of the event.
        Returns:
            A :class:`asyncio.Task` or :class:`asyncio.Future` that will
            complete once all event handlers have been called.
        Raises:
            :class:`TypeError` if an invalid event is supplied.
        '''

        if not self._dispatcher_running:
            return

        event_cls = event
        if not isinstance(event, Event):
            if not isinstance(event, type) or not issubclass(event, Event):
                raise TypeError("events must be a subclass or instance of Event")
            # create an instance of the event if passed a class
            event = event(**kw)
        else:
            event_cls = event.__class__

        if id(self) in event._delivered_to:
            return
        event._delivered_to.add(id(self))

        handlers = set()
        for cls in event._parent_event_classes():
            for handler in self._dispatch_handlers[cls.event_name]:
                if event._is_filtered(handler.f):
                    continue

                if getattr(handler.f, '_oneshot_handler', False):
                    # Disable oneshot events prior to actual dispatch
                    handler.disable()
                handlers.add(handler)

        return asyncio.ensure_future(self._dispatch_event(event, handlers), loop=self._loop)

    async def _dispatch_event(self, event, handlers):
        # iterate through events from child->parent
        # update the dispatched_to set for each event so each handler
        # only receives the most specific event if they are monitoring for both.

        try:
            # dispatch to local handlers
            for handler in handlers:
                if isinstance(handler.f, asyncio.Future):
                    event._dispatch_to_future(handler.f)
                else:
                    result = event._dispatch_to_func(handler.f)
                    if asyncio.iscoroutine(result):
                        await result

            # dispatch to children
            for child in self._dispatch_children:
                child.dispatch_event(event)

            # dispatch to self methods
            result = event._dispatch_to_obj(self)
            if asyncio.iscoroutine(result):
                await result

            # dispatch to parent dispatcher
            if self._dispatch_parent:
                self._dispatch_parent.dispatch_event(event)

        except exceptions.StopPropogation:
            pass

    def _abort_event_futures(self, exc):
        '''Sets an exception on all pending Future handlers

        This prevents coroutines awaiting a Future from blocking forever
        should a hard failure occur with the connection.
        '''
        handlers = set()
        for evh in self._dispatch_handlers.values():
            for h in evh:
                handlers.add(h)

        for handler in handlers:
            if isinstance(handler.f, asyncio.Future):
                if not handler.f.done():
                    handler.f.set_exception(exc)
                handler.disable()

    async def wait_for(self, event_or_filter, timeout=30):
        '''Waits for the specified event to be sent to the current object.

        Args:
            event_or_filter (:class:`Event`): Either a :class:`Event` class
                or a :class:`Filter` instance to wait to trigger
            timeout: Maximum time to wait for the event.  Pass None to wait indefinitely.
        Returns:
            The :class:`Event` instance that was dispatched
        Raises:
            :class:`asyncio.TimeoutError`
        '''
        f = asyncio.Future(loop=self._loop) # replace with loop.create_future in 3.5.2
        # TODO: add a timer that logs every 5 seconds that the event is still being
        # waited on.  Will help novice programmers realize why their program is hanging.
        f = oneshot(f)

        if isinstance(event_or_filter, Filter):
            f = filter_handler(event_or_filter)(f)
            event = event_or_filter._event
        else:
            event = event_or_filter

        self.add_event_handler(event, f)
        if timeout:
             return await asyncio.wait_for(f, timeout, loop=self._loop)
        return await f


def oneshot(f):
    '''Event handler decorator; causes the handler to only be dispatched to once.'''
    f._oneshot_handler = True
    return f


def filter_handler(event, **filters):
    '''Decorates a handler function or Future to only be called if a filter is matched.

    A handler may apply multiple separate filters; the handlers will be called
    if any of those filters matches.

    For example::

        # Handle only if the majorWin animation completed
        @filter_handler(cozmo.anim.EvtAnimationCompleted, animation_name="majorWin")

        # Handle only when the observed object is a LightCube
        @filter_handler(cozmo.objects.EvtObjectObserved, obj=lambda obj: isinstance(cozmo.objects.LightCube))

    Args:
        event (:class:`Event`): The event class to match on
        filters (dict): Zero or more event parameters to filter on.  Values may
            be either strings for exact matches, or functions which accept the
            value as the first argument and return a bool indicating whether
            the value passes the filter.
    '''

    if isinstance(event, Filter):
        if len(filters) != 0:
            raise ValueError("Cannot supply filter values when passing a Filter as the first argument")
        filter = event
    else:
        filter = Filter(event, **filters)

    def filter_property(f):
        if hasattr(f, '_handler_filters'):
            f._handler_filters.append(filter)
        else:
            f._handler_filters = [filter]
        return f
    return filter_property


class Filter:
    """Provides fine-grain filtering of events for dispatch.

    See the ::func::`filter_handler` method for further details.
    """

    def __init__(self, event, **filters):
        if not issubclass(event, Event):
            raise TypeError("event must be a subclass of Event (not an instance)")
        self._event = event
        self._filters = filters
        for key in self._filters.keys():
            if not hasattr(event, key):
                raise AttributeError("Event %s does not define property %s", event.__name__, key)

    def __setattr__(self, key, val):
        if key[0] == '_':
            return super().__setattr__(key, val)
        if not hasattr(self._event, key):
                raise AttributeError("Event %s does not define property %s", self._event.__name__, key)
        self._filters[key] = val

    def __call__(self, evt):
        for prop, filter in self._filters.items():
            val = getattr(evt, prop)
            if callable(filter):
                if not filter(val):
                    return False
            elif val != filter:
                return False
        return True


async def wait_for_first(*futures, discard_remaining=True, loop=None):
    '''Wait the first of a set of futures to complete.

    Eg::

        event = cozmo.event.wait_for_first(
            coz.world.wait_for_new_cube(),
            playing_anim.wait_for(cozmo.anim.EvtAnimationCompleted)
        )

    If more than one completes during a single event loop run, then
    if any of those results are not exception, one of them will be selected
    (at random, as determined by ``set.pop``) to be returned, else one
    of the result exceptions will be raised instead.

    Args:
        futures (list of :class:`asyncio.Future`): The futures or coroutines to wait on.
        discard_remaining (bool): Cancel or discard the results of the futures
            that did not return first.
        loop (:class:`asyncio.BaseEventLoop`): The event loop to wait on.
    Returns:
        The first result, or raised exception
    '''
    done, pending = await asyncio.wait(futures, loop=loop, return_when=asyncio.FIRST_COMPLETED)

    # collect the results from all "done" futures; only one will be returned
    result = None
    for fut in done:
        try:
            fut_result = fut.result()
            if result is None or isinstance(result, BaseException):
                result = fut_result
        except Exception as exc:
            if result is None:
                result = exc

    if discard_remaining:
        # cancel the pending futures
        for fut in pending:
            fut.cancel()

    if isinstance(result, BaseException):
        raise result
    return result


def _abort_futures(exc):
    '''Trigger the exception handler for all pending Future handlers.'''
    for obj in active_dispatchers:
        obj._abort_event_futures(exc)



# Copyright (c) 2016 Anki, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License in the file LICENSE.txt or at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = ('USBMuxError', 'ProtocolError', 'DeviceNotConnected',
    'ConnectionRefused', 'ConnectionFailed', 'QueueNotifyCM', 'USBMux',
    'connect_to_usbmux')

import asyncio
import collections
import contextlib
import plistlib
import socket
import struct
import sys
import time


DEFAULT_SOCKET_PATH = '/var/run/usbmuxd'
DEFAULT_SOCKET_PORT = 27015
DEFAULT_MAX_WAIT = 2

PLIST_VERSION = 1

ACTION_ATTACHED = 'attached'
ACTION_DETACHED = 'detached'


class USBMuxError(Exception): pass

class ProtocolError(USBMuxError): pass

class DeviceNotConnected(USBMuxError): pass

class ConnectionRefused(USBMuxError): pass

class ConnectionFailed(USBMuxError): pass


class PlistProto(asyncio.Protocol):
    def connection_made(self, transport):
        self.transport = transport
        self._buf = bytearray()

    def data_received(self, data):
        self._buf += data
        while len(self._buf) > 4:
            length = struct.unpack('I', self._buf[:4])[0]
            if len(self._buf) < length:
                return
            ver, req, tag = struct.unpack('III', self._buf[4:16])
            if ver != PLIST_VERSION:
                raise ProtocolError("Unsupported protocol version from usbmux stream")
            pldata = plistlib.loads(self._buf[16:length])
            self.msg_received(pldata)
            self._buf = self._buf[length:]

    def send_msg(self, **kw):
        pl = plistlib.dumps(kw)
        self.transport.write(struct.pack('IIII', len(pl) + 16, 1, 8, 1))
        self.transport.write(pl)

    def msg_received(self, msg):
        '''Called when a plist record is received'''


class USBMuxConnector(PlistProto):
    '''Opens a connection to a port on a device'''

    def __init__(self, device_id, port, waiter):
        self.device_id = device_id
        self.port = port
        self.waiter = waiter

    def connection_made(self, transport):
        super().connection_made(transport)

        self.send_msg(
                MessageType='Connect',
                ClientVersionString='pyusbmux',
                ProgName='pyusbmux',
                DeviceID=self.device_id,
                PortNumber=socket.htons(self.port)
            )

    def connection_lost(self, exc):
        if self.waiter.done():
            return
        self.waiter.set_exception(exc)

    def msg_received(self, msg):
        if msg['MessageType'] != 'Result':
            return
        status = msg['Number']
        if status == 0:
            self.waiter.set_result(None)

            # ensure no futher data is received until the protocol
            # is switched to the application protocol.
            self.transport.pause_reading()

        elif status == 2:
            self.waiter.set_exception(DeviceNotConnected("Device %s is not currently connected" % (self.device_id,)))

        elif status == 3:
            self.waiter.set_exception(ConnectionRefused("Connection refused to device_id=%s port=%d" % (self.device_id, self.port)))

        else:
            self.waiter.set_exception(ConnectionFailed("Protocol error connecting to device %s" % (self.device_id,)))


class _ProtoSwitcher(asyncio.Protocol):
    def __init__(self, loop, initial_protocol):
        self._loop = loop
        self._transport = None
        self.protocol = initial_protocol

    def switch_protocol(self, protocol_factory):
        self.protocol = protocol_factory()
        if self._transport:
            self._loop.call_soon(self.protocol.connection_made, self._transport)
            self._loop.call_soon(self._transport.resume_reading)
        return self.protocol

    def connection_made(self, transport):
        self._transport = transport
        self.protocol.connection_made(transport)

    def connection_lost(self, exc):
        self.protocol.connection_lost(exc)

    def pause_writing(self):
        self.protocol.pause_writing()

    def resume_writing(self):
        self.protocol.resume_writing()

    def data_received(self, data):
        self.protocol.data_received(data)

    def eof_received(self):
        self.protocol.eof_received()


class USBMux(PlistProto):
    '''USBMux wraps a connection to the USBMux daemon.

    Use ``connect_to_usbmux`` or call ``connect`` on an instance of this
    class to connect to the daemon.

    Once connected, the ``attached`` attribute is populated with a dictionary
    keyed by an integer device id, and with a dictionary of values about the
    connected device.

    The ``attached`` dictionary is populated asynchronously and may be empty
    after ``connect`` returns.

    Subclasses of USBMux will have their ``device_attached`` and
    ``device_detached`` methods called as devices are made available through
    the connected mux.

    Alternatively call :meth:`wait_for_attach` to wait for a new device to
    be made available, or use the :meth:`attach_watcher` method to obtain
    a context manager to iterate over all devices as they connect and
    disconect.

    The ``connect`` call will open a TCP connection to a specific port
    on a specific device.  :meth:`connect_to_first` can be used if it doesn't
    matter which device is connected to, as long as the requested port is open.
    '''
    def __init__(self, loop, mux_socket_path=DEFAULT_SOCKET_PATH, mux_socket_port=DEFAULT_SOCKET_PORT):
        #: Currently attached devices, keyed by integer device_id
        self.attached = {}
        self.loop = loop
        self.mux_socket_path = mux_socket_path
        self.mux_socket_port = mux_socket_port
        self._attach_notify = QueueNotify(loop=loop)

    async def _connect_transport(self, protocol_factory):
        if sys.platform in ('win32', 'cygwin'):
            return await self.loop.create_connection(protocol_factory, host='127.0.0.1', port=self.mux_socket_port)
        else:
            result = await self.loop.create_unix_connection(protocol_factory, self.mux_socket_path)
            return result

    async def connect(self):
        '''Opens a connection to the USBMux daemon on the local machine.

        :func:`connect_to_usbmux` provides a convenient wrapper to this method.
        '''
        self._waiter = asyncio.Future(loop=self.loop)
        await self._connect_transport(lambda: self)
        await self._waiter

    def connection_made(self, transport):
        super().connection_made(transport)

        self.send_msg(
                MessageType='Listen',
                ClientVersionString='pyusbmux',
                ProgName='pyusbmux'
            )

    def connection_lost(self, exc):
        super().connection_lost(exc)

        if not self._waiter.done():
            self._waiter.set_exception(exc)

    def msg_received(self, msg):
        mt = msg.get('MessageType')

        if mt == 'Result':
            if msg['Number'] == 0:
                self._waiter.set_result(None)
            else:
                self._waiter.set_exception(ConnectionFailed())

        elif mt == 'Attached':
            device_id = msg['Properties']['DeviceID']
            self.attached[device_id] = msg['Properties']
            self.device_attached(device_id, msg['Properties'])
            self._attach_notify.notify((ACTION_ATTACHED, device_id, msg['Properties']))

        elif mt == 'Detached':
            device_id = msg['DeviceID']
            if device_id in self.attached:
                props = self.attached[device_id]
                del(self.attached[device_id])
                self._attach_notify.notify((ACTION_DETACHED, device_id, props))
            self.device_detached(device_id)

    async def connect_to_device(self, protocol_factory, device_id, port):
        '''Open a TCP connection to a port on a device.

        Args:
            protocol_factory (callable): A callable that returns an asyncio.Protocol implementation
            device_id (int): The id of the device to connect to
            port (int): The port to connect to on the target device
        Returns:
            (dict, asyncio.Transport, asyncio.Protocol): The device information,
                connected transport and protocol.
        Raises:
            A USBMuxError subclass instance such as ConnectionRefused
        '''
        waiter = asyncio.Future(loop=self.loop)
        connector = USBMuxConnector(device_id, port, waiter)
        transport, switcher = await self._connect_transport(lambda: _ProtoSwitcher(self.loop, connector))

        # wait for the connection to succeed or fail
        await waiter

        app_protocol = switcher.switch_protocol(protocol_factory)
        device_info = self.attached.get(device_id) or {}
        return device_info, transport, app_protocol

    async def wait_for_serial(self, serial, timeout=DEFAULT_MAX_WAIT):
        '''Wait for a device with the specified serial number to attach.

        Args:
            serial (string): Serial number of the device to wait for.
            timeout (float): The maximum amount of time in seconds to wait for a
                matching device to be connected.
                Set to None to wait indefinitely, or -1 to only check currently
                connected devices.
        Returns:
            int: The device id of the connected device
        Raises:
            asyncio.TimeoutError if the device with the specified serial number doesn't appear.
        '''
        timeout = Timeout(timeout)
        with self.attach_watcher(include_existing=True) as watcher:
            while not timeout.expired:
                action, device_id, info = await watcher.wait_for_next(timeout.remaining)
                if action != ACTION_ATTACHED:
                    continue
                if info['SerialNumber'].lower() == serial.lower():
                    return device_id

        raise asyncio.TimeoutError("No devices matching serial number found")

    async def connect_to_first_device(self, protocol_factory, port, timeout=DEFAULT_MAX_WAIT,
            include=None, exclude=None):
        '''Open a TCP connection to the first device that has the requested port open.

        Args:
            protocol_factory (callable): A callable that returns an asyncio.Protocol implementation.
            port (int): The port to connect to on the target device.
            timeout (float): The maximum amount of time to wait for a suitable device to be connected.
        Returns:
            (dict, asyncio.Transport, asyncio.Protocol): The device information,
                connected transport and protocol.
        Raises:
            asyncio.TimeoutError if no devices with the requested port become
                available in the specified time.
        '''
        with self.attach_watcher(include_existing=True) as watcher:
            timeout = Timeout(timeout)
            while not timeout.expired:
                action, device_id, info = await watcher.wait_for_next(timeout.remaining)
                if action != ACTION_ATTACHED:
                    continue
                if exclude is not None and device_id in exclude:
                    continue
                if include is not None and device_id not in include:
                    continue
                try:
                    return await self.connect_to_device(protocol_factory, device_id, port)
                except USBMuxError:
                    pass

        raise asyncio.TimeoutError("No available devices")

    async def wait_for_attach(self, timeout=None):
        '''Wait for the next device attachment event.

        Args:
            timeout (float): Maximum amount of time to wait for an event, or None for no timeout
        Returns:
            int: The device id that attached.
        Raises:
            asyncio.TimeoutError if no devices with the requested port become
                available in the specified time.
        '''
        timeout = Timeout(timeout)
        with self.attach_watcher() as watcher:
            while True:
                action, device_id, info = await watcher.wait_for_next(timeout.remaining)
                if action == ACTION_ATTACHED:
                    return device_id

    def attach_watcher(self, include_existing=False):
        '''Returns a context manager that will record and make available all attach/detach notifications.

        The context manager yields events consisting of (action, device_id, device_info) tuples,
        where ``action`` is either :const:`ACTION_ATTACHED` or :const:`ACTION_DETACHED`
        and ``device_info`` is a dictionary of information specific to that device,
        such as the serial number.

        Args:
            include_existing (bool): If True then a stream of fake attached events
                will be generated for all existing connected devices ahead of
                monitoring for newly attached devices.
        Returns:
            :class:`QueueNotifyCM`
        '''
        initial_data = None
        if include_existing:
            initial_data = [(ACTION_ATTACHED, device_id, info)
                    for (device_id, info) in self.attached.items()]
        return self._attach_notify.get_contextmanager(initial_data=initial_data)

    def device_attached(self, device_id, properties):
        pass

    def device_detached(self, device_id):
        pass


async def connect_to_usbmux(mux_socket_path=DEFAULT_SOCKET_PATH, mux_socket_port=DEFAULT_SOCKET_PORT, loop=None):
    '''Connect to a USBMux endpoint.

    Args:
        mux_socket_path (string) - The path of the Unix socket of the mux daemon (used on non Windows platforms)
        mux_socket_port (int) - The TCP port number of the mux daemon (used on Windows platforms)
        loop (asyncio.BaseLoop) - Event loop to connect on; defaults to the current active event loop
    Returns:
        USBMux instance
    Raises:
        Exception on connection refused or other error
    '''
    if loop is None:
        loop = asyncio.get_event_loop()

    mux = USBMux(loop, mux_socket_path=mux_socket_path, mux_socket_port=mux_socket_port)
    await mux.connect()
    return mux


class QueueNotify:
    '''Provide a context manager to queue and read asynchronous notifications.

    While the context manager is active, all notifications are queued and
    read by calling ``wait_for_next`` on the returned QueueNotifyCM
    object.  If none are available, then the method will wait for the specified
    amount of time for a new entry to arrive.

    Multiple context managers can be active concurrently receiving the same
    notifications.
    '''
    def __init__(self, loop=None):
        self.loop = loop
        self._active = set()

    def notify(self, value):
        for entry in self._active:
            entry._notify(value)

    def get_contextmanager(self, initial_data=None, max_qsize=None):
        ctx = QueueNotifyCM(self, initial_data=initial_data, max_qsize=max_qsize, loop=self.loop)
        self._active.add(ctx)
        return ctx

    def context_done(self, ctx):
        self._active.discard(ctx)


class QueueNotifyCM:
    '''Helper class for QueueNotify.'''
    def __init__(self, mgr, initial_data=None, max_qsize=None, loop=None):
        self.loop = loop
        self._mgr = mgr
        self._wake = None
        if initial_data is None:
            initial_data = []
        self._q = collections.deque(initial_data, max_qsize)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._mgr.context_done(self)
        return False

    def _notify(self, item):
        self._q.append(item)
        if self._wake is not None and not self._wake.done():
            self._wake.set_result(True)
            self._wake = None

    async def wait_for_next(self, timeout=None):
        '''Wait for the next available notification.

        Will return immediately if entries are already waiting to be read,
        else wait up to ``timeout`` seconds for a new entry to arrive.
        '''
        try:
            return self._q.popleft()
        except IndexError:
            pass
        self._wake = asyncio.Future(loop=self.loop)
        await asyncio.wait_for(self._wake, loop=self.loop, timeout=timeout)
        return self._q.popleft()


class Timeout:
    '''Helper class to track timeout state.'''
    def __init__(self, timeout=None):
        self.timeout = timeout
        self.start = time.time()

    @property
    def remaining(self):
        if self.timeout is None:
            return None
        return self.timeout - (time.time() - self.start)

    @property
    def expired(self):
        return self.timeout is not None and self.remaining <= 0



# Copyright (c) 2016 Anki, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License in the file LICENSE.txt or at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''Support for Cozmo's camera.

Cozmo has a built-in camera which he uses to observe the world around him.

The :class:`Camera` class defined in this module is made available as
:attr:`cozmo.world.World.camera` and can be used to enable/disable image
sending, enable/disable color images, modify various camera settings,
read the robot's unique camera calibration settings, as well as observe
raw unprocessed images being sent by the robot.

Generally, however, it is more useful to observe
:class:`cozmo.world.EvtNewCameraImage` events, which include the raw camera
images along with annotated images, which can illustrate objects the robot
has identified.
'''

# __all__ should order by constants, event classes, other classes, functions.
__all__ = ['EvtNewRawCameraImage', 'CameraConfig', 'Camera']

import functools
import io

_img_processing_available = True

try:
    import numpy as np
    from PIL import Image
except ImportError as exc:
    np = None
    _img_processing_available = exc


from . import event
from . import logger
from . import util

from ._clad import _clad_to_engine_iface, _clad_to_engine_cozmo, _clad_to_game_cozmo


_clad_res = _clad_to_game_cozmo.ImageResolution
RESOLUTIONS = {
    _clad_res.VerificationSnapshot: (16, 16),
    _clad_res.QQQQVGA: (40, 30),
    _clad_res.QQQVGA: (80, 60),
    _clad_res.QQVGA: (160, 120),
    _clad_res.QVGA: (320, 240),
    _clad_res.CVGA: (400, 296),
    _clad_res.VGA: (640, 480),
    _clad_res.SVGA: (800, 600),
    _clad_res.XGA: (1024, 768),
    _clad_res.SXGA: (1280, 960),
    _clad_res.UXGA: (1600, 1200),
    _clad_res.QXGA: (2048, 1536),
    _clad_res.QUXGA: (3200, 2400)
}


# wrap functions/methods that require NumPy or PIL with this
# decorator to ensure they fail with a useful error if those packages
# are not loaded.
def _require_img_processing(f):
    @functools.wraps(f)
    def wrapper(*a, **kw):
        if _img_processing_available is not True:
            raise ImportError("Camera image processing not available: %s" % _img_processing_available)
        return f(*a, **kw)
    return wrapper


class EvtNewRawCameraImage(event.Event):
    '''Dispatched when a new raw image is received from the robot's camera.

    See also :class:`~cozmo.world.EvtNewCameraImage` which provides access
    to both the raw image and a scaled and annotated version.
    '''
    image = 'A PIL.Image.Image object'


class CameraConfig:
    """The fixed properties for Cozmo's Camera

    A full 3x3 calibration matrix for doing 3D reasoning based on the camera
    images would look like:

        +--------------+--------------+---------------+
        |focal_length.x|      0       |    center.x   |
        +--------------+--------------+---------------+
        |       0      |focal_length.y|    center.y   |
        +--------------+--------------+---------------+
        |       0      |       0      |        1      |
        +--------------+--------------+---------------+
    """

    def __init__(self,
                 focal_length_x: float,
                 focal_length_y: float,
                 center_x: float,
                 center_y: float,
                 fov_x_degrees: float,
                 fov_y_degrees: float,
                 min_exposure_time_ms: int,
                 max_exposure_time_ms: int,
                 min_gain: float,
                 max_gain: float):
        self._focal_length = util.Vector2(focal_length_x, focal_length_y)
        self._center = util.Vector2(center_x, center_y)
        self._fov_x = util.degrees(fov_x_degrees)
        self._fov_y = util.degrees(fov_y_degrees)
        self._min_exposure_time_ms = min_exposure_time_ms
        self._max_exposure_time_ms = max_exposure_time_ms
        self._min_gain = min_gain
        self._max_gain = max_gain

    @classmethod
    def _create_from_clad(cls, cs):
        return cls(cs.focalLengthX, cs.focalLengthY,
                   cs.centerX, cs.centerY,
                   cs.fovX, cs.fovY,
                   cs.minCameraExposureTime_ms, cs.maxCameraExposureTime_ms,
                   cs.minCameraGain, cs.maxCameraGain)

    # Fixed camera properties (calibrated for each robot at the factory).

    @property
    def focal_length(self):
        ''':class:`cozmo.util.Vector2`: The focal length of the camera.

        This is focal length combined with pixel skew (as the pixels aren't
        perfectly square), so there are subtly different values for x and y.
        It is in floating point pixel values e.g. <288.87, 288.36>.
        '''
        return self._focal_length

    @property
    def center(self):
        ''':class:`cozmo.util.Vector2`: The focal center of the camera.

        This is the position of the optical center of projection within the
        image. It will be close to the center of the image, but adjusted based
        on the calibration of the lens at the factory. It is in floating point
        pixel values e.g. <155.11, 111.40>.
        '''
        return self._center

    @property
    def fov_x(self):
        ''':class:`cozmo.util.Angle`: The x (horizontal) field of view.'''
        return self._fov_x

    @property
    def fov_y(self):
        ''':class:`cozmo.util.Angle`: The y (vertical) field of view.'''
        return self._fov_y

    # The fixed range of values supported for this camera.

    @property
    def min_exposure_time_ms(self):
        '''int: The minimum supported exposure time in milliseconds.'''
        return self._min_exposure_time_ms

    @property
    def max_exposure_time_ms(self):
        '''int: The maximum supported exposure time in milliseconds.'''
        return self._max_exposure_time_ms

    @property
    def min_gain(self):
        '''float: The minimum supported camera gain.'''
        return self._min_gain

    @property
    def max_gain(self):
        '''float: The maximum supported camera gain.'''
        return self._max_gain


class Camera(event.Dispatcher):
    '''Represents Cozmo's camera.

    The Camera object receives images from Cozmo's camera and emits
    EvtNewRawCameraImage events.

    The :class:`cozmo.world.World` instance observes the camera and provides
    more useful methods for accessing the camera images.

    .. important::
        The camera will not receive any image data unless you
        explicitly enable it by setting :attr:`Camera.image_stream_enabled`
        to ``True``
    '''

    def __init__(self, robot, **kw):
        super().__init__(**kw)
        self.robot = robot
        self._image_stream_enabled = None
        self._color_image_enabled = None
        self._camera_config = None  # type: CameraConfig
        self._gain = 0.0
        self._exposure_ms = 0
        self._auto_exposure_enabled = True

        if np is None:
            logger.warning("Camera image processing not available due to missng NumPy or Pillow packages: %s" % _img_processing_available)
        else:
            # set property to ensure clad initialization is sent.
            self.image_stream_enabled = False
            self.color_image_enabled = False
        self._reset_partial_state()

    def enable_auto_exposure(self):
        '''Enable auto exposure on Cozmo's Camera.

        Enable auto exposure on Cozmo's camera to constantly update the exposure
        time and gain values based on the recent images. This is the default mode
        when any SDK program starts.
        '''
        msg = _clad_to_engine_iface.SetCameraSettings(enableAutoExposure=True)
        self.robot.conn.send_msg(msg)

    def set_manual_exposure(self, exposure_ms, gain):
        '''Set manual exposure values for Cozmo's Camera.

        Disable auto exposure on Cozmo's camera and force the specified exposure
        time and gain values.

        Args:
            exposure_ms (int): The desired exposure time in milliseconds.
                Must be within the robot's
                :attr:`~cozmo.camera.Camera.config` exposure range from
                :attr:`~cozmo.camera.CameraConfig.min_exposure_time_ms` to
                :attr:`~cozmo.camera.CameraConfig.max_exposure_time_ms`
            gain (float): The desired gain value.
                Must be within the robot's
                :attr:`~cozmo.camera.Camera.camera_config` gain range from
                :attr:`~cozmo.camera.CameraConfig.min_gain` to
                :attr:`~cozmo.camera.CameraConfig.max_gain`

        Raises:
            :class:`ValueError` if supplied an out-of-range exposure or gain.
        '''
        cam = self.config

        if (exposure_ms < cam.min_exposure_time_ms) or (exposure_ms > cam.max_exposure_time_ms):
            raise ValueError('exposure_ms %s out of range %s..%s' %
                             (exposure_ms, cam.min_exposure_time_ms, cam.max_exposure_time_ms))

        if (gain < cam.min_gain) or (gain > cam.max_gain):
            raise ValueError('gain %s out of range %s..%s' %
                             (gain, cam.min_gain, cam.max_gain))

        msg = _clad_to_engine_iface.SetCameraSettings(enableAutoExposure=False,
                                                      exposure_ms=exposure_ms,
                                                      gain=gain)
        self.robot.conn.send_msg(msg)

    #### Private Methods ####

    def _reset_partial_state(self):
        self._partial_data = None
        self._partial_image_id = None
        self._partial_invalid = False
        self._partial_size = 0
        self._partial_metadata = None
        self._last_chunk_id = -1

    def _set_config(self, clad_config):
        self._config = CameraConfig._create_from_clad(clad_config)

    #### Properties ####

    @property
    @_require_img_processing
    def image_stream_enabled(self):
        '''bool: Set to true to receive camera images from the robot.'''
        if np is None:
            return False

        return self._image_stream_enabled

    @image_stream_enabled.setter
    @_require_img_processing
    def image_stream_enabled(self, enabled):
        if self._image_stream_enabled == enabled:
            return

        self._image_stream_enabled = enabled

        if enabled:
            image_send_mode = _clad_to_engine_cozmo.ImageSendMode.Stream
        else:
            image_send_mode = _clad_to_engine_cozmo.ImageSendMode.Off

        msg = _clad_to_engine_iface.ImageRequest(
                robotID=self.robot.robot_id, mode=image_send_mode)

        self.robot.conn.send_msg(msg)

    @property
    @_require_img_processing
    def color_image_enabled(self):
        '''bool: Set to true to receive color images from the robot.'''
        if np is None:
            return False

        return self._color_image_enabled

    @color_image_enabled.setter
    @_require_img_processing
    def color_image_enabled(self, enabled):
        if self._color_image_enabled == enabled:
            return

        self._color_image_enabled = enabled

        msg = _clad_to_engine_iface.EnableColorImages(enable = enabled)
        self.robot.conn.send_msg(msg)

    @property
    def config(self):
        ''':class:`cozmo.camera.CameraConfig`: The read-only config/calibration for the camera'''
        return self._config

    @property
    def is_auto_exposure_enabled(self):
        '''bool: True if auto exposure is currently enabled

        If auto exposure is enabled the `gain` and `exposure_ms`
        values will constantly be updated by Cozmo.
        '''
        return self._auto_exposure_enabled

    @property
    def gain(self):
        '''float: The current camera gain setting.'''
        return self._gain

    @property
    def exposure_ms(self):
        '''int: The current camera exposure setting in milliseconds.'''
        return self._exposure_ms

    #### Private Event Handlers ####

    def _recv_msg_image_chunk(self, evt, *, msg):
        if np is None:
            return
        if self._partial_image_id is not None and msg.chunkId == 0:
            if not self._partial_invalid:
                logger.debug("Lost final chunk of image; discarding")
            self._partial_image_id = None

        if self._partial_image_id is None:
            if msg.chunkId != 0:
                if not self._partial_invalid:
                    logger.debug("Received chunk of broken image")
                self._partial_invalid = True
                return
            # discard any previous in-progress image
            self._reset_partial_state()
            self._partial_image_id = msg.imageId
            self._partial_metadata = msg

            max_size = msg.imageChunkCount * _clad_to_game_cozmo.ImageConstants.IMAGE_CHUNK_SIZE
            width, height = RESOLUTIONS[msg.resolution]
            max_size = width * height * 3 # 3 bytes (RGB) per pixel
            self._partial_data = np.empty(max_size, dtype=np.uint8)

        if msg.chunkId != (self._last_chunk_id + 1) or msg.imageId != self._partial_image_id:
            logger.debug("Image missing chunks; discarding (last_chunk_id=%d partial_image_id=%s)",
                    self._last_chunk_id, self._partial_image_id)
            self._reset_partial_state()
            self._partial_invalid = True
            return

        offset = self._partial_size
        self._partial_data[offset:offset+len(msg.data)] = msg.data
        self._partial_size += len(msg.data)
        self._last_chunk_id = msg.chunkId

        if msg.chunkId == (msg.imageChunkCount - 1):
            self._process_completed_image()
            self._reset_partial_state()

    def _recv_msg_current_camera_params(self, evt, *, msg):
        self._gain = msg.cameraGain
        self._exposure_ms = msg.exposure_ms
        self._auto_exposure_enabled = msg.autoExposureEnabled

    def _process_completed_image(self):
        data = self._partial_data[0:self._partial_size]
        
        # The first byte of the image is whether or not it is in color
        is_color_image = data[0] != 0
        
        if self._partial_metadata.imageEncoding ==  _clad_to_game_cozmo.ImageEncoding.JPEGMinimizedGray:
            width, height = RESOLUTIONS[self._partial_metadata.resolution]
            
            if is_color_image:
                # Color images are half width
                width = width // 2
                data = _minicolor_to_jpeg(data, width, height)
            else:
                data = _minigray_to_jpeg(data, width, height)
                
        image = Image.open(io.BytesIO(data)).convert('RGB')

        # Color images need to be resized to the proper resolution
        if is_color_image:
            size = RESOLUTIONS[self._partial_metadata.resolution]
            image = image.resize(size)
        
        self._latest_image = image
        self.dispatch_event(EvtNewRawCameraImage, image=image)


    #### Public Event Handlers ####


@_require_img_processing
def _minigray_to_jpeg(minigray, width, height):
        "Converts miniGrayToJpeg format to normal jpeg format"
        #This should be 'exactly' what is done in the miniGrayToJpeg function in encodedImage.cpp
        header50 = np.array([
              0xFF, 0xD8, 0xFF, 0xE0, 0x00, 0x10, 0x4A, 0x46, 0x49, 0x46, 0x00, 0x01, 0x01, 0x00, 0x00, 0x01,
              0x00, 0x01, 0x00, 0x00, 0xFF, 0xDB, 0x00, 0x43, 0x00, 0x10, 0x0B, 0x0C, 0x0E, 0x0C, 0x0A, 0x10, #// 0x19 = QTable
              0x0E, 0x0D, 0x0E, 0x12, 0x11, 0x10, 0x13, 0x18, 0x28, 0x1A, 0x18, 0x16, 0x16, 0x18, 0x31, 0x23,
              0x25, 0x1D, 0x28, 0x3A, 0x33, 0x3D, 0x3C, 0x39, 0x33, 0x38, 0x37, 0x40, 0x48, 0x5C, 0x4E, 0x40,
              0x44, 0x57, 0x45, 0x37, 0x38, 0x50, 0x6D, 0x51, 0x57, 0x5F, 0x62, 0x67, 0x68, 0x67, 0x3E, 0x4D,

              #//0x71, 0x79, 0x70, 0x64, 0x78, 0x5C, 0x65, 0x67, 0x63, 0xFF, 0xC0, 0x00, 0x0B, 0x08, 0x00, 0xF0, #// 0x5E = Height x Width
              0x71, 0x79, 0x70, 0x64, 0x78, 0x5C, 0x65, 0x67, 0x63, 0xFF, 0xC0, 0x00, 0x0B, 0x08, 0x01, 0x28, #// 0x5E = Height x Width

              #//0x01, 0x40, 0x01, 0x01, 0x11, 0x00, 0xFF, 0xC4, 0x00, 0xD2, 0x00, 0x00, 0x01, 0x05, 0x01, 0x01,
              0x01, 0x90, 0x01, 0x01, 0x11, 0x00, 0xFF, 0xC4, 0x00, 0xD2, 0x00, 0x00, 0x01, 0x05, 0x01, 0x01,

              0x01, 0x01, 0x01, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01, 0x02, 0x03, 0x04,
              0x05, 0x06, 0x07, 0x08, 0x09, 0x0A, 0x0B, 0x10, 0x00, 0x02, 0x01, 0x03, 0x03, 0x02, 0x04, 0x03,
              0x05, 0x05, 0x04, 0x04, 0x00, 0x00, 0x01, 0x7D, 0x01, 0x02, 0x03, 0x00, 0x04, 0x11, 0x05, 0x12,
              0x21, 0x31, 0x41, 0x06, 0x13, 0x51, 0x61, 0x07, 0x22, 0x71, 0x14, 0x32, 0x81, 0x91, 0xA1, 0x08,
              0x23, 0x42, 0xB1, 0xC1, 0x15, 0x52, 0xD1, 0xF0, 0x24, 0x33, 0x62, 0x72, 0x82, 0x09, 0x0A, 0x16,
              0x17, 0x18, 0x19, 0x1A, 0x25, 0x26, 0x27, 0x28, 0x29, 0x2A, 0x34, 0x35, 0x36, 0x37, 0x38, 0x39,
              0x3A, 0x43, 0x44, 0x45, 0x46, 0x47, 0x48, 0x49, 0x4A, 0x53, 0x54, 0x55, 0x56, 0x57, 0x58, 0x59,
              0x5A, 0x63, 0x64, 0x65, 0x66, 0x67, 0x68, 0x69, 0x6A, 0x73, 0x74, 0x75, 0x76, 0x77, 0x78, 0x79,
              0x7A, 0x83, 0x84, 0x85, 0x86, 0x87, 0x88, 0x89, 0x8A, 0x92, 0x93, 0x94, 0x95, 0x96, 0x97, 0x98,
              0x99, 0x9A, 0xA2, 0xA3, 0xA4, 0xA5, 0xA6, 0xA7, 0xA8, 0xA9, 0xAA, 0xB2, 0xB3, 0xB4, 0xB5, 0xB6,
              0xB7, 0xB8, 0xB9, 0xBA, 0xC2, 0xC3, 0xC4, 0xC5, 0xC6, 0xC7, 0xC8, 0xC9, 0xCA, 0xD2, 0xD3, 0xD4,
              0xD5, 0xD6, 0xD7, 0xD8, 0xD9, 0xDA, 0xE1, 0xE2, 0xE3, 0xE4, 0xE5, 0xE6, 0xE7, 0xE8, 0xE9, 0xEA,
              0xF1, 0xF2, 0xF3, 0xF4, 0xF5, 0xF6, 0xF7, 0xF8, 0xF9, 0xFA, 0xFF, 0xDA, 0x00, 0x08, 0x01, 0x01,
              0x00, 0x00, 0x3F, 0x00
            ], dtype=np.uint8)
        
        return _mini_to_jpeg_helper(minigray, width, height, header50)

@_require_img_processing
def _minicolor_to_jpeg(minicolor, width, height):
        "Converts miniColorToJpeg format to normal jpeg format"
        #This should be 'exactly' what is done in the miniColorToJpeg function in encodedImage.cpp
        header = np.array([
             0xFF, 0xD8, 0xFF, 0xE0, 0x00, 0x10, 0x4A, 0x46, 0x49, 0x46, 0x00, 0x01, 0x01, 0x00, 0x00, 0x01,
             0x00, 0x01, 0x00, 0x00, 0xFF, 0xDB, 0x00, 0x43, 0x00, 0x10, 0x0B, 0x0C, 0x0E, 0x0C, 0x0A, 0x10, # 0x19 = QTable
             0x0E, 0x0D, 0x0E, 0x12, 0x11, 0x10, 0x13, 0x18, 0x28, 0x1A, 0x18, 0x16, 0x16, 0x18, 0x31, 0x23,
             0x25, 0x1D, 0x28, 0x3A, 0x33, 0x3D, 0x3C, 0x39, 0x33, 0x38, 0x37, 0x40, 0x48, 0x5C, 0x4E, 0x40,
             0x44, 0x57, 0x45, 0x37, 0x38, 0x50, 0x6D, 0x51, 0x57, 0x5F, 0x62, 0x67, 0x68, 0x67, 0x3E, 0x4D,
             0x71, 0x79, 0x70, 0x64, 0x78, 0x5C, 0x65, 0x67, 0x63, 0xFF, 0xC0, 0x00, 17, # 8+3*components
             0x08, 0x00, 0xF0, # 0x5E = Height x Width
             0x01, 0x40,
             0x03, # 3 components
             0x01, 0x21, 0x00, # Y 2x1 res
             0x02, 0x11, 0x00, # Cb
             0x03, 0x11, 0x00, # Cr
             0xFF, 0xC4, 0x00, 0xD2, 0x00, 0x00, 0x01, 0x05, 0x01, 0x01,
             0x01, 0x01, 0x01, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01, 0x02, 0x03, 0x04,
             0x05, 0x06, 0x07, 0x08, 0x09, 0x0A, 0x0B, 0x10, 0x00, 0x02, 0x01, 0x03, 0x03, 0x02, 0x04, 0x03,
             0x05, 0x05, 0x04, 0x04, 0x00, 0x00, 0x01, 0x7D, 0x01, 0x02, 0x03, 0x00, 0x04, 0x11, 0x05, 0x12,
             0x21, 0x31, 0x41, 0x06, 0x13, 0x51, 0x61, 0x07, 0x22, 0x71, 0x14, 0x32, 0x81, 0x91, 0xA1, 0x08,
             0x23, 0x42, 0xB1, 0xC1, 0x15, 0x52, 0xD1, 0xF0, 0x24, 0x33, 0x62, 0x72, 0x82, 0x09, 0x0A, 0x16,
             0x17, 0x18, 0x19, 0x1A, 0x25, 0x26, 0x27, 0x28, 0x29, 0x2A, 0x34, 0x35, 0x36, 0x37, 0x38, 0x39,
             0x3A, 0x43, 0x44, 0x45, 0x46, 0x47, 0x48, 0x49, 0x4A, 0x53, 0x54, 0x55, 0x56, 0x57, 0x58, 0x59,
             0x5A, 0x63, 0x64, 0x65, 0x66, 0x67, 0x68, 0x69, 0x6A, 0x73, 0x74, 0x75, 0x76, 0x77, 0x78, 0x79,
             0x7A, 0x83, 0x84, 0x85, 0x86, 0x87, 0x88, 0x89, 0x8A, 0x92, 0x93, 0x94, 0x95, 0x96, 0x97, 0x98,
             0x99, 0x9A, 0xA2, 0xA3, 0xA4, 0xA5, 0xA6, 0xA7, 0xA8, 0xA9, 0xAA, 0xB2, 0xB3, 0xB4, 0xB5, 0xB6,
             0xB7, 0xB8, 0xB9, 0xBA, 0xC2, 0xC3, 0xC4, 0xC5, 0xC6, 0xC7, 0xC8, 0xC9, 0xCA, 0xD2, 0xD3, 0xD4,
             0xD5, 0xD6, 0xD7, 0xD8, 0xD9, 0xDA, 0xE1, 0xE2, 0xE3, 0xE4, 0xE5, 0xE6, 0xE7, 0xE8, 0xE9, 0xEA,
             0xF1, 0xF2, 0xF3, 0xF4, 0xF5, 0xF6, 0xF7, 0xF8, 0xF9, 0xFA,
             0xFF, 0xDA, 0x00, 12,
             0x03, # 3 components
             0x01, 0x00, # Y
             0x02, 0x00, # Cb same AC/DC
             0x03, 0x00, # Cr same AC/DC
             0x00, 0x3F, 0x00
            ], dtype=np.uint8)

        return _mini_to_jpeg_helper(minicolor, width, height, header)

@_require_img_processing
def _mini_to_jpeg_helper(mini, width, height, header):
        bufferIn = mini.tolist()
        currLen = len(mini)

        headerLength = len(header)
        # For worst case expansion
        bufferOut = np.array([0] * (currLen*2 + headerLength), dtype=np.uint8)

        for i in range(headerLength):
            bufferOut[i] = header[i]

        bufferOut[0x5e] = height >> 8
        bufferOut[0x5f] = height & 0xff
        bufferOut[0x60] = width  >> 8
        bufferOut[0x61] = width  & 0xff
        # Remove padding at the end
        while (bufferIn[currLen-1] == 0xff):
            currLen -= 1

        off = headerLength
        for i in range(currLen-1):
            bufferOut[off] = bufferIn[i+1]
            off += 1
            if (bufferIn[i+1] == 0xff):
                bufferOut[off] = 0
                off += 1

        bufferOut[off] = 0xff
        off += 1
        bufferOut[off] = 0xD9

        bufferOut[:off]
        return np.asarray(bufferOut)



# Copyright (c) 2016-2017 Anki, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License in the file LICENSE.txt or at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''This module provides a simple GUI viewer for Cozmo's camera.

It uses Tkinter, the standard Python GUI toolkit which is optionally available
on most platforms, and also depends on the Pillow and numpy libraries for
image processing.


See the online SDK documentation for details on how to install these extra
packages on your platform.

The easiest way to make use of this viewer is to call
:func:`cozmo.run.connect_with_tkviewer`.

Warning:
    This package requires Python to have Tkinter installed to display the GUI.
'''


# __all__ should order by constants, event classes, other classes, functions.
__all__ = ['TkImageViewer']

import cozmo
import collections
import functools
import queue
import platform
import time

from PIL import Image, ImageDraw, ImageTk
import tkinter

from . import world


class TkThreadable:
    '''A mixin for adding threadsafe calls to tkinter methods.'''
    def __init__(self, *a, **kw):
        self._thread_queue = queue.Queue()
        self.after(50, self._thread_call_dispatch)

    def call_threadsafe(self, method, *a, **kw):
        self._thread_queue.put((method, a, kw))

    def _thread_call_dispatch(self):
        while True:
            try:
                method, a, kw = self._thread_queue.get(block=False)
                self.after_idle(method, *a, **kw)
            except queue.Empty:
                break
        self.after(50, self._thread_call_dispatch)


class TkImageViewer(tkinter.Frame, TkThreadable):
    '''Simple Tkinter camera viewer.'''

    # TODO: rewrite this whole thing.  Make a generic camera widget
    # that can be used in other Tk applications.  Also handle resizing
    # the window properly.
    def __init__(self,
            tk_root=None, refresh_interval=10, image_scale = 2,
            window_name = "CozmoView", force_on_top=True):
        if tk_root is None:
            tk_root = tkinter.Tk()
        tkinter.Frame.__init__(self, tk_root)
        TkThreadable.__init__(self)
        self._img_queue = collections.deque(maxlen=1)

        self._refresh_interval = refresh_interval
        self.scale = image_scale
        self.width = None
        self.height = None

        self.tk_root = tk_root
        tk_root.wm_title(window_name)

        self.label  = tkinter.Label(self.tk_root,image=None)
        self.tk_root.protocol("WM_DELETE_WINDOW", self._delete_window)
        self._isRunning = True
        self.robot = None
        self.handler = None
        self._first_image = True
        tk_root.aspect(4,3,4,3)

        if force_on_top:
            # force window on top of all others, regardless of focus
            tk_root.wm_attributes("-topmost", 1)

        self.last_configure = time.time()
        self.tk_root.bind("<Configure>", self.configure)
        self._repeat_draw_frame()

    async def connect(self, coz_conn):
        self.robot = await coz_conn.wait_for_robot()
        self.robot.camera.image_stream_enabled = True
        self.handler = self.robot.world.add_event_handler(
            world.EvtNewCameraImage, self.image_event)

    def disconnect(self):
        if self.handler:
            self.handler.disable()
        self.call_threadsafe(self.quit)

    def configure(self, event):
        # hack to interrupt feedback loop between image resizing
        # and frame resize detection; there has to be a better solution to this.
        if time.time() - self.last_configure < 0.1:
            return
        if event.width < 50 or event.height < 50:
            return
        self.last_configure = time.time()
        self.height = event.height
        self.width = event.width


    def image_event(self, evt, *, image, **kw):
        if self._first_image or self.width is None:
            img = image.annotate_image(scale=self.scale)
        else:
            img = image.annotate_image(fit_size=(self.width, self.height))
        self._img_queue.append(img)

    def _delete_window(self):
        self.tk_root.destroy()
        self.quit()
        self._isRunning = False

    def _draw_frame(self):
        if ImageTk is None:
            return

        try:
            image = self._img_queue.popleft()
        except IndexError:
            # no new image
            return

        self._first_image = False
        photoImage = ImageTk.PhotoImage(image)

        self.label.configure(image=photoImage)
        self.label.image = photoImage
        self.label.pack()

    def _repeat_draw_frame(self, event=None):
        self._draw_frame()
        self.after(self._refresh_interval, self._repeat_draw_frame)



# Copyright (c) 2016 Anki, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License in the file LICENSE.txt or at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''Helper routines for dealing with Cozmo's lights and colors.'''

# __all__ should order by constants, event classes, other classes, functions.
__all__ = ['green', 'red', 'blue', 'white', 'off',
           'green_light', 'red_light', 'blue_light', 'white_light', 'off_light',
           'Color', 'Light']

import copy

from . import logger


class Color:
    '''A Color to be used with a Light.

    Either int_color or rgb may be used to specify the actual color.
    Any alpha components (from int_color) are ignored - all colors are fully opaque.

    Args:
        int_color (int): A 32 bit value holding the binary RGBA value (where A
            is ignored and forced to be fully opaque).
        rgb (tuple): A tuple holding the integer values from 0-255 for (reg, green, blue)
        name (str): A name to assign to this color
    '''

    def __init__(self, int_color=None, rgb=None, name=None):
        self.name = name
        self._int_color = 0
        if int_color is not None:
            self._int_color = int_color | 0xff
        elif rgb is not None:
            self._int_color = (rgb[0] << 24) | (rgb[1] << 16) | (rgb[2] << 8) | 0xff

    @property
    def int_color(self):
        '''int: The encoded integer value of the color.'''
        return self._int_color


#: :class:`Color`: Green color instance.
green = Color(name="green", int_color=0x00ff00ff)

#: :class:`Color`: Red color instance.
red = Color(name="red", int_color=0xff0000ff)

#: :class:`Color`: Blue color instance.
blue = Color(name="blue", int_color=0x0000ffff)

#: :class:`Color`: White color instance.
white = Color(name="white", int_color=0xffffffff)

#: :class:`Color`:  instance representing no color (LEDs off).
off = Color(name="off")


class Light:
    '''Lights are used with LightCubes and Cozmo's backpack.

    Lights may either be "on" or "off", though in practice any colors may be
    assigned to either state (including no color/light).
    '''

    def __init__(self, on_color=off, off_color=off, on_period_ms=250,
            off_period_ms=0, transition_on_period_ms=0, transition_off_period_ms=0):
        self._on_color = on_color
        self._off_color = off_color
        self._on_period_ms = on_period_ms
        self._off_period_ms = off_period_ms
        self._transition_on_period_ms = transition_on_period_ms
        self._transition_off_period_ms = transition_off_period_ms

    @property
    def on_color(self):
        ''':class:`Color`: The Color shown when the light is on.'''
        return self._on_color

    @on_color.setter
    def on_color(self, color):
        if not isinstance(color, Color):
            raise TypeError("Must specify a Color")
        self._on_color = color

    @property
    def off_color(self):
        ''':class:`Color`: The Color shown when the light is off.'''
        return self._off_color

    @off_color.setter
    def off_color(self, color):
        if not isinstance(color, Color):
            raise TypeError("Must specify a Color")
        self._off_color = color

    @property
    def on_period_ms(self):
        '''int: The number of milliseconds the light should be "on" for for each cycle.'''
        return self._on_period_ms

    @on_period_ms.setter
    def on_period_ms(self, ms):
        if not 0 < ms < 2**32:
            raise ValueError("Invalid value")
        self._on_period_ms = ms

    @property
    def off_period_ms(self):
        '''int: The number of milliseconds the light should be "off" for for each cycle.'''
        return self._off_period_ms

    @off_period_ms.setter
    def off_period_ms(self, ms):
        if not 0 < ms < 2**32:
            raise ValueError("Invalid value")
        self._off_period_ms = ms

    @property
    def transition_on_period_ms(self):
        '''int: The number of milliseconds to take to transition the light to the on color.'''
        return self._transition_on_period_ms

    @transition_on_period_ms.setter
    def transition_on_period_ms(self, ms):
        if not 0 < ms < 2**32:
            raise ValueError("Invalid value")
        self._transition_on_period_ms = ms

    @property
    def transition_off_period_ms(self):
        '''int: The number of milliseconds to take to transition the light to the off color.'''
        return self._transition_off_period_ms

    @transition_off_period_ms.setter
    def transition_off_period_ms(self, ms):
        if not 0 < ms < 2**32:
            raise ValueError("Invalid value")
        self._transition_off_period_ms = ms

    def flash(self, on_period_ms=250, off_period_ms=250, off_color=off):
        '''Convenience function to make a flashing version of an existing Light instance.

        Args:
            on_period_ms (int): The number of milliseconds the light should be "on" for for each cycle.
            off_period_ms (int): The number of milliseconds the light should be "off" for for each cycle.
            off_color (:class:`Color`): The color to flash to for the off state.
        Returns:
            :class:`Color` instance.
        '''
        flasher = copy.copy(self)
        flasher.on_period_ms = on_period_ms
        flasher.off_period_ms = off_period_ms
        flasher.off_color=off_color
        return flasher

def _set_light(msg, idx, light):
    # For use with clad light messages specifically.
    if not isinstance(light, Light):
        raise TypeError("Expected a lights.Light")
    msg.onColor[idx] = light.on_color.int_color
    msg.offColor[idx] = light.off_color.int_color
    msg.onPeriod_ms[idx] = light.on_period_ms
    msg.offPeriod_ms[idx] = light.off_period_ms
    msg.transitionOnPeriod_ms[idx] = light.transition_on_period_ms
    msg.transitionOffPeriod_ms[idx] = light.transition_off_period_ms

#There is a glitch so it will always flash unless on_color==off_color
#ticket is COZMO-3319

#: :class:`Light`: A steady green colored LED light.
green_light = Light(on_color=green, off_color=green)

#: :class:`Light`: A steady red colored LED light.
red_light = Light(on_color=red, off_color=red)

#: :class:`Light`: A steady blue colored LED light.
blue_light = Light(on_color=blue, off_color=blue)

#: :class:`Light`: A steady white colored LED light.
white_light = Light(on_color=white, off_color=white)

#: :class:`Light`: A steady off (non-illuminated LED light).
off_light = Light(on_color=off, off_color=off)



# Copyright (c) 2016-2017 Anki, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License in the file LICENSE.txt or at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''Utility classes and functions.'''


# __all__ should order by constants, event classes, other classes, functions.
# (util keeps class related functions close to their associated class)
__all__ = ['Angle', 'degrees', 'radians',
           'ImageBox',
           'Distance', 'distance_mm', 'distance_inches',
           'Pose', 'pose_quaternion', 'pose_z_angle',
           'Position',
           'Rotation', 'rotation_quaternion', 'rotation_z_angle',
           'angle_z_to_quaternion',
           'Speed', 'speed_mmps',
           'Timeout', 'Vector2', 'Vector3']


import collections
import math
import re
import time
from ._clad import _clad_to_engine_anki


class ImageBox(collections.namedtuple('ImageBox', 'top_left_x top_left_y width height')):
    '''Defines a bounding box within an image frame.

    This is used when objects, faces and pets are observed to denote where in
    the robot's camera view the object, face or pet actually appears.  It's then
    used by the :mod:`cozmo.annotate` module to show an outline of a box around
    the object, face or pet.
    '''

    @classmethod
    def _create_from_clad_rect(cls, img_rect):
        return cls(img_rect.x_topLeft, img_rect.y_topLeft,
                   img_rect.width, img_rect.height)

    def __mul__(self, other):
        return ImageBox(self[0] * other, self[1] * other, self[2] * other, self[3] * other)


class Angle:
    '''Represents an angle.

    Use the :func:`degrees` or :func:`radians` convenience methods to generate
    an Angle instance.

    Args:
        radians (float): The number of radians the angle should represent
            (cannot be combined with ``degrees``)
        degrees (float): The number of degress the angle should represent
            (cannot be combined with ``radians``)
    '''

    __slots__ = ('_radians')

    def __init__(self, radians=None, degrees=None):
        if radians is None and degrees is None:
            raise ValueError("Expected either the degrees or radians keyword argument")
        if radians and degrees:
            raise ValueError("Expected either the degrees or radians keyword argument, not both")

        if degrees is not None:
            radians = degrees * math.pi / 180
        self._radians = radians

    def __repr__(self):
        return "<%s %.2f radians (%.2f degrees)>" % (self.__class__.__name__, self.radians, self.degrees)

    def __add__(self, other):
        if not isinstance(other, Angle):
            raise TypeError("Unsupported operand for + expected Angle")
        return radians(self.radians + other.radians)

    def __sub__(self, other):
        if not isinstance(other, Angle):
            raise TypeError("Unsupported operand for - expected Angle")
        return radians(self.radians - other.radians)

    def __mul__(self, other):
        if not isinstance(other, (int, float)):
            raise TypeError("Unsupported operand for * expected number")
        return radians(self.radians * other)

    def __truediv__(self, other):
        if not isinstance(other, (int, float)):
            raise TypeError("Unsupported operand for / expected number")
        return radians(self.radians / other)

    def _cmp_int(self, other):
        if not isinstance(other, Angle):
            raise TypeError("Unsupported operand for < expected Angle")
        return (self.radians > other.radians) - (self.radians < other.radians)

    def __eq__(self, other):
        return self._cmp_int(other) == 0

    def __ne__(self, other):
        return self._cmp_int(other) != 0

    def __gt__(self, other):
        return self._cmp_int(other) > 0

    def __lt__(self, other):
        return self._cmp_int(other) < 0

    def __ge__(self, other):
        return self._cmp_int(other) >= 0

    def __le__(self, other):
        return self._cmp_int(other) <= 0

    @property
    def radians(self):
        '''float: The angle in radians.'''
        return self._radians

    @property
    def degrees(self):
        '''float: The angle in degrees.'''
        return self._radians / math.pi * 180


def degrees(degrees):
    '''Returns an :class:`cozmo.util.Angle` instance set to the specified number of degrees.'''
    return Angle(degrees=degrees)


def radians(radians):
    '''Returns an :class:`cozmo.util.Angle` instance set to the specified number of radians.'''
    return Angle(radians=radians)


class Distance:
    '''Represents a distance.

    The class allows distances to be returned in either millimeters or inches.

    Use the :func:`distance_inches` or :func:`distance_mm` convenience methods to generate
    a Distance instance.

    Args:
        distance_mm (float): The number of millimeters the distance should
            represent (cannot be combined with ``distance_inches``).
        distance_inches (float): The number of inches the distance should
            represent (cannot be combined with ``distance_mm``).
    '''

    __slots__ = ('_distance_mm')

    def __init__(self, distance_mm=None, distance_inches=None):
        if distance_mm is None and distance_inches is None:
            raise ValueError("Expected either the distance_mm or distance_inches keyword argument")
        if distance_mm and distance_inches:
            raise ValueError("Expected either the distance_mm or distance_inches keyword argument, not both")

        if distance_inches is not None:
            distance_mm = distance_inches * 25.4
        self._distance_mm = distance_mm

    def __repr__(self):
        return "<%s %.2f mm (%.2f inches)>" % (self.__class__.__name__, self.distance_mm, self.distance_inches)

    def __add__(self, other):
        if not isinstance(other, Distance):
            raise TypeError("Unsupported operand for + expected Distance")
        return distance_mm(self.distance_mm + other.distance_mm)

    def __sub__(self, other):
        if not isinstance(other, Distance):
            raise TypeError("Unsupported operand for - expected Distance")
        return distance_mm(self.distance_mm - other.distance_mm)

    def __mul__(self, other):
        if not isinstance(other, (int, float)):
            raise TypeError("Unsupported operand for * expected number")
        return distance_mm(self.distance_mm * other)

    def __truediv__(self, other):
        if not isinstance(other, (int, float)):
            raise TypeError("Unsupported operand for / expected number")
        return distance_mm(self.distance_mm / other)

    @property
    def distance_mm(self):
        '''float: The distance in millimeters'''
        return self._distance_mm

    @property
    def distance_inches(self):
        '''float: The distance in inches'''
        return self._distance_mm / 25.4


def distance_mm(distance_mm):
    '''Returns an :class:`cozmo.util.Distance` instance set to the specified number of millimeters.'''
    return Distance(distance_mm=distance_mm)


def distance_inches(distance_inches):
    '''Returns an :class:`cozmo.util.Distance` instance set to the specified number of inches.'''
    return Distance(distance_inches=distance_inches)


class Speed:
    '''Represents a speed.

    This class allows speeds to be measured  in millimeters per second.

    Use :func:`speed_mmps` convenience methods to generate
    a Speed instance.

    Args:
        speed_mmps (float): The number of millimeters per second the speed
            should represent.
    '''

    __slots__ = ('_speed_mmps')

    def __init__(self, speed_mmps=None):
        if speed_mmps is None:
            raise ValueError("Expected speed_mmps keyword argument")
        self._speed_mmps = speed_mmps

    def __repr__(self):
        return "<%s %.2f mmps>" % (self.__class__.__name__, self.speed_mmps)

    def __add__(self, other):
        if not isinstance(other, Speed):
            raise TypeError("Unsupported operand for + expected Speed")
        return speed_mmps(self.speed_mmps + other.speed_mmps)

    def __sub__(self, other):
        if not isinstance(other, Speed):
            raise TypeError("Unsupported operand for - expected Speed")
        return speed_mmps(self.speed_mmps - other.speed_mmps)

    def __mul__(self, other):
        if not isinstance(other, (int, float)):
            raise TypeError("Unsupported operand for * expected number")
        return speed_mmps(self.speed_mmps * other)

    def __truediv__(self, other):
        if not isinstance(other, (int, float)):
            raise TypeError("Unsupported operand for / expected number")
        return speed_mmps(self.speed_mmps / other)

    @property
    def speed_mmps(self):
        '''float: The speed in millimeters per second (mmps).'''
        return self._speed_mmps


def speed_mmps(speed_mmps):
    '''Returns an :class:`cozmo.util.Speed` instance set to the specified millimeters per second speed'''
    return Speed(speed_mmps=speed_mmps)


class Pose:
    '''Represents where an object is in the world.

    Use the :func:'pose_quaternion' to return pose in the form of
    position and rotation defined by a quaternion

    Use the :func:'pose_z_angle' to return pose in the form of
    position and rotation defined by rotation about the z axis

    When the engine is initialized, and whenever Cozmo is de-localized (i.e.
    whenever Cozmo no longer knows where he is - e.g. when he's picked up)
    Cozmo creates a new pose starting at (0,0,0) with no rotation, with
    origin_id incremented to show that these poses cannot be compared with
    earlier ones. As Cozmo drives around, his pose (and the pose of other
    objects he observes - e.g. faces, cubes etc.) is relative to this initial
    position and orientation.

    The coordinate space is relative to Cozmo, where Cozmo's origin is the
    point on the ground between Cozmo's two front wheels:

    The X axis is Cozmo's forward direction
    The Y axis is to Cozmo's left
    The Z axis is up

    Only poses of the same origin_id can safely be compared or operated on
    '''

    __slots__ = ('_position', '_rotation', '_origin_id')

    def __init__(self, x, y, z, q0=None, q1=None, q2=None, q3=None, angle_z=None, origin_id=0):
        self._position = Position(x,y,z)
        self._rotation = Rotation(q0,q1,q2,q3,angle_z)
        self._origin_id = origin_id

    @classmethod
    def _create_from_clad(cls, pose):
        return cls(pose.x, pose.y, pose.z,
                   q0=pose.q0, q1=pose.q1, q2=pose.q2, q3=pose.q3,
                   origin_id=pose.originID)

    def __repr__(self):
        return "<%s %s %s origin_id=%d>" % (self.__class__.__name__, self.position, self.rotation, self.origin_id)

    def __add__(self, other):
        if not isinstance(other, Pose):
            raise TypeError("Unsupported operand for + expected Pose")
        pos = self.position + other.position
        rot = self.rotation + other.rotation
        return pose_quaternion(pos.x, pos.y, pos.z, rot.q0, rot.q1, rot.q2, rot.q3)

    def __sub__(self, other):
        if not isinstance(other, Pose):
            raise TypeError("Unsupported operand for - expected Pose")
        pos = self.position - other.position
        rot = self.rotation - other.rotation
        return pose_quaternion(pos.x, pos.y, pos.z, rot.q0, rot.q1, rot.q2, rot.q3)

    def __mul__(self, other):
        if not isinstance(other, (int, float)):
            raise TypeError("Unsupported operand for * expected number")
        pos = self.position * other
        rot = self.rotation * other
        return pose_quaternion(pos.x, pos.y, pos.z, rot.q0, rot.q1, rot.q2, rot.q3)

    def __truediv__(self, other):
        if not isinstance(other, (int, float)):
            raise TypeError("Unsupported operand for / expected number")
        pos = self.position / other
        rot = self.rotation / other
        return pose_quaternion(pos.x, pos.y, pos.z, rot.q0, rot.q1, rot.q2, rot.q3)

    def define_pose_relative_this(self, new_pose):
        '''Creates a new pose such that new_pose's origin is now at the location of this pose.

        Args:
            new_pose (:class:`cozmo.util.Pose`): The pose which origin is being changed.
        Returns:
            A :class:`cozmo.util.pose` object for which the origin was this pose's origin.
        '''

        if not isinstance(new_pose, Pose):
            raise TypeError("Unsupported type for new_origin, must be of type Pose")
        x,y,z = self.position.x_y_z
        angle_z = self.rotation.angle_z
        new_x,new_y,new_z = new_pose.position.x_y_z
        new_angle_z = new_pose.rotation.angle_z

        res_x = x + math.cos(angle_z.radians)*new_x - math.sin(angle_z.radians)*new_y
        res_y = y + math.sin(angle_z.radians)*new_x + math.cos(angle_z.radians)*new_y
        res_z = z + new_z
        res_angle = angle_z + new_angle_z
        return Pose(res_x, res_y, res_z, angle_z=res_angle)

    def encode_pose(self):
        x, y, z = self.position.x_y_z
        q0, q1, q2, q3 = self.rotation.q0_q1_q2_q3
        return _clad_to_engine_anki.PoseStruct3d(x, y, z, q0, q1, q2, q3, self.origin_id)

    def invalidate(self):
        '''Mark this pose as being invalid (unusable)'''
        self._origin_id = -1

    def is_comparable(self, other_pose):
        '''Are these two poses comparable.

        Poses are comparable if they're valid and having matching origin IDs.

        Args:
            other_pose (:class:`cozmo.util.Pose`): The other pose to compare against.
        Returns:
            bool: True if the two poses are comparable, False otherwise.
        '''
        return (self.is_valid and other_pose.is_valid and
                (self.origin_id == other_pose.origin_id))

    @property
    def is_valid(self):
        '''bool: Returns True if this is a valid, usable pose.'''
        return self.origin_id >= 0

    @property
    def position(self):
        ''':class:`cozmo.util.Position`: The position component of this pose.'''
        return self._position

    @property
    def rotation(self):
        ''':class:`cozmo.util.Rotation`: The rotation component of this pose.'''
        return self._rotation

    @property
    def origin_id(self):
        '''int: An ID maintained by the engine which represents which coordinate frame this pose is in.'''
        return self._origin_id

    @origin_id.setter
    def origin_id(self, value):
        '''Allows this to be changed later in case it was not originally defined.'''
        if not isinstance(value, int):
            raise TypeError("The type of origin_id must be int")
        self._origin_id = value


def pose_quaternion(x, y, z, q0, q1, q2, q3, origin_id=0):
    '''Returns a :class:`cozmo.util.Pose` instance set to the pose given in quaternion format.'''
    return Pose(x, y, z, q0=q0, q1=q1, q2=q2, q3=q3, origin_id=origin_id)

def pose_z_angle(x, y, z, angle_z, origin_id=0):
    '''Returns a :class:`cozmo.util.Pose` instance set to the pose given in z angle format.'''
    return Pose(x, y, z, angle_z=angle_z, origin_id=origin_id)


class Rotation:
    '''Represents the rotation of an object in the world. Can be generated with
    quaternion to define its rotation in 3d space, or with only a z axis rotation
    to define things limited to the x-y plane like Cozmo.

    Use the :func:`rotation_quaternion` to return rotation defined by a quaternion.

    Use the :func:`rotation_angle_z` to return rotation defined by an angle in the z axis.
    '''
    __slots__ = ('_q0', '_q1', '_q2', '_q3')

    def __init__(self, q0=None, q1=None, q2=None, q3=None, angle_z=None):
        is_quaternion = not (q0 is None) and not (q1 is None) and not (q2 is None) and not (q3 is None)

        if not is_quaternion and angle_z is None:
            raise ValueError("Expected either the q0 q1 q2 and q3 or angle_z keyword arguments")
        if is_quaternion and angle_z:
            raise ValueError("Expected either the q0 q1 q2 and q3 or angle_z keyword argument, not both")
        if angle_z is not None:
            if not isinstance(angle_z, Angle):
                raise TypeError("Unsupported type for angle_z expected Angle")
            q0,q1,q2,q3 = angle_z_to_quaternion(angle_z)

        self._q0, self._q1, self._q2, self._q3 = q0, q1, q2, q3

    def __repr__(self):
        return ("<%s q0: %.2f q1: %.2f q2: %.2f q3: %.2f (angle_z: %s)>" %
            (self.__class__.__name__, self.q0, self.q1, self.q2, self.q3, self.angle_z))

    #These are only for angle_z because quaternion addition/subtraction is not relevant here
    def __add__(self, other):
        if not isinstance(other, Rotation):
            raise TypeError("Unsupported operand for + expected Rotation")
        return rotation_z_angle(self.angle_z + other.angle_z)

    def __sub__(self, other):
        if not isinstance(other, Rotation):
            raise TypeError("Unsupported operand for - expected Rotation")
        return rotation_z_angle(self.angle_z - other.angle_z)

    def __mul__(self, other):
        if not isinstance(other, (int,float)):
            raise TypeError("Unsupported operand for * expected number")
        return rotation_z_angle(self.angle_z * other)

    def __truediv__(self, other):
        if not isinstance(other, (int,float)):
            raise TypeError("Unsupported operand for / expected number")
        return rotation_z_angle(self.angle_z / other)

    @property
    def q0(self):
        '''float: The q0 (w) value of the quaternion defining this rotation.'''
        return self._q0

    @property
    def q1(self):
        '''float: The q1 (i) value of the quaternion defining this rotation.'''
        return self._q1

    @property
    def q2(self):
        '''float: The q2 (j) value of the quaternion defining this rotation.'''
        return self._q2

    @property
    def q3(self):
        '''float: The q3 (k) value of the quaternion defining this rotation.'''
        return self._q3

    @property
    def q0_q1_q2_q3(self):
        '''tuple of float: Contains all elements of the quaternion (q0,q1,q2,q3)'''
        return self._q0,self._q1,self._q2,self._q3

    @property
    def angle_z(self):
        '''class:`Angle`: The z Euler component of the object's rotation. Defined as the rotation in the z axis.'''
        q0,q1,q2,q3 = self.q0_q1_q2_q3
        return Angle(radians=math.atan2(2*(q1*q2+q0*q3), 1-2*(q2**2+q3**2)))


def rotation_quaternion(q0, q1, q2, q3):
    '''Returns a :class:`Rotation` instance set by a quaternion.'''
    return Rotation(q0=q0, q1=q1, q2=q2, q3=q3)


def rotation_z_angle(angle_z):
    '''Returns a class:`Rotation` instance set by an angle in the z axis'''
    return Rotation(angle_z=angle_z)

def angle_z_to_quaternion(angle_z):
    '''This function converts an angle in the z axis (Euler angle z component) to a quaternion.

    Args:
        angle_z (:class:`cozmo.util.Angle`): The z axis angle.

    Returns:
        q0,q1,q2,q3 (float, float, float, float): A tuple with all the members
            of a quaternion defined by angle_z.
    '''

    #Define the quaternion to be converted from a Euler angle (x,y,z) of 0,0,angle_z
    #These equations have their original equations above, and simplified implemented
    # q0 = cos(x/2)*cos(y/2)*cos(z/2) + sin(x/2)*sin(y/2)*sin(z/2)
    q0 = math.cos(angle_z.radians/2)
    # q1 = sin(x/2)*cos(y/2)*cos(z/2) - cos(x/2)*sin(y/2)*sin(z/2)
    q1 = 0
    # q2 = cos(x/2)*sin(y/2)*cos(z/2) + sin(x/2)*cos(y/2)*sin(z/2)
    q2 = 0
    # q3 = cos(x/2)*cos(y/2)*sin(z/2) - sin(x/2)*sin(y/2)*cos(z/2)
    q3 = math.sin(angle_z.radians/2)
    return q0,q1,q2,q3


class Vector2:
    '''Represents a 2D Vector (type/units aren't specified)

    Args:
        x (float): X component
        y (float): Y component
    '''

    __slots__ = ('_x', '_y')

    def __init__(self, x, y):
        self._x = x
        self._y = y

    @property
    def x(self):
        '''float: The x component.'''
        return self._x

    @property
    def y(self):
        '''float: The y component.'''
        return self._y

    @property
    def x_y(self):
        '''tuple (float, float): The X, Y elements of the Vector2 (x,y)'''
        return self._x, self._y

    def __repr__(self):
        return "<%s x: %.2f y: %.2f>" % (self.__class__.__name__, self.x, self.y)

    def __add__(self, other):
        if not isinstance(other, Vector2):
            raise TypeError("Unsupported operand for + expected Vector2")
        return Vector2(self.x + other.x, self.y + other.y)

    def __sub__(self, other):
        if not isinstance(other, Vector2):
            raise TypeError("Unsupported operand for - expected Vector2")
        return Vector2(self.x - other.x, self.y - other.y)

    def __mul__(self, other):
        if not isinstance(other, (int, float)):
            raise TypeError("Unsupported operand for * expected number")
        return Vector2(self.x * other, self.y * other)

    def __truediv__(self, other):
        if not isinstance(other, (int, float)):
            raise TypeError("Unsupported operand for / expected number")
        return Vector2(self.x / other, self.y / other)


class Vector3:
    '''Represents a 3D Vector (type/units aren't specified)

    Args:
        x (float): X component
        y (float): Y component
        z (float): Z component
    '''

    __slots__ = ('_x', '_y', '_z')

    def __init__(self, x, y, z):
        self._x = x
        self._y = y
        self._z = z

    @property
    def x(self):
        '''float: The x component.'''
        return self._x

    @property
    def y(self):
        '''float: The y component.'''
        return self._y

    @property
    def z(self):
        '''float: The z component.'''
        return self._z

    @property
    def x_y_z(self):
        '''tuple (float, float, float): The X, Y, Z elements of the Vector3 (x,y,z)'''
        return self._x, self._y, self._z

    def __repr__(self):
        return "<%s x: %.2f y: %.2f z: %.2f>" % (self.__class__.__name__, self.x, self.y, self.z)

    def __add__(self, other):
        if not isinstance(other, Vector3):
            raise TypeError("Unsupported operand for + expected Vector3")
        return Vector3(self.x + other.x, self.y + other.y, self.z + other.z)

    def __sub__(self, other):
        if not isinstance(other, Vector3):
            raise TypeError("Unsupported operand for - expected Vector3")
        return Vector3(self.x - other.x, self.y - other.y, self.z - other.z)

    def __mul__(self, other):
        if not isinstance(other, (int, float)):
            raise TypeError("Unsupported operand for * expected number")
        return Vector3(self.x * other, self.y * other, self.z * other)

    def __truediv__(self, other):
        if not isinstance(other, (int, float)):
            raise TypeError("Unsupported operand for / expected number")
        return Vector3(self.x / other, self.y / other, self.z / other)


class Position(Vector3):
    '''Represents the position of an object in the world.

    A position consists of its x, y and z values in millimeters.

    Args:
        x (float): X position in millimeters
        y (float): Y position in millimeters
        z (float): Z position in millimeters
    '''


class Timeout:
    '''Utility class to keep track of a timeout condition.

    This measures a timeout from the point in time that the class
    is instantiated.

    Args:
        timeout (float): Amount of time (in seconds) allotted to pass before
            considering the timeout condition to be met.
        use_inf (bool): If True, then :attr:`remaining` will return
            :const:`math.inf` if `timeout` is None, else it will return
            `None`.
    '''
    def __init__(self, timeout=None, use_inf=False):
        self.start = time.time()
        self.timeout = timeout
        self.use_inf = use_inf

    @property
    def is_timed_out(self):
        '''bool: True if the timeout has expired.'''
        if self.timeout is None:
            return False
        return time.time() - self.start > self.timeout

    @property
    def remaining(self):
        '''float: The number of seconds remaining before reaching the timeout.

        Will return a number of zero or higher, even if the timer has
        since expired (it will never return a negative value).

        Will return None or math.inf (if ``use_inf`` was passed as ``True``
        to the constructor) if the original timeout was ``None``.
        '''
        if self.timeout is None:
            return math.inf if self.use_inf else None
        return max(0, self.timeout - (time.time() - self.start))



# Copyright (c) 2016-2017 Anki, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License in the file LICENSE.txt or at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''Object and Power Cube recognition.

Cozmo can recognize and track a number of different types of objects.

These objects may be visible (currently observed by the robot's camera)
and tappable (in the case of the Power Cubes that ship with the robot).

Power Cubes are known as a :class:`LightCube` by the SDK.  Each cube has
controllable lights, and sensors that can determine when its being moved
or tapped.

Objects can emit several events such as :class:`EvtObjectObserved` when
the robot sees (or continues to see) the object with its camera, or
:class:`EvtObjectTapped` if a power cube is tapped by a player.  You
can either observe the object's instance directly, or capture all such events
for all objects by observing them on :class:`cozmo.world.World` instead.

All observable objects have a marker attached to them, which allows Cozmo
to recognize the object and it's position and rotation("pose").  You can attach
markers to your own objects for Cozmo to recognize by printing them out from the
online documentation.  They will be detected as :class:`CustomObject` instances.
'''


# __all__ should order by constants, event classes, other classes, functions.
__all__ = ['LightCube1Id', 'LightCube2Id', 'LightCube3Id', 'OBJECT_VISIBILITY_TIMEOUT',
           'EvtObjectAppeared', 'EvtObjectAvailable',
           'EvtObjectMoving', 'EvtObjectMovingStarted', 'EvtObjectMovingStopped',
           'EvtObjectTapped',
           'EvtObjectConnectChanged', 'EvtObjectDisappeared', 'EvtObjectObserved',
           'ObservableElement', 'ObservableObject', 'LightCube', 'Charger',
           'CustomObject', 'CustomObjectMarkers', 'CustomObjectTypes', 'FixedCustomObject']


import collections
import math
import time

from . import logger

from . import action
from . import event
from . import lights
from . import util

from ._clad import _clad_to_engine_iface, _clad_to_game_cozmo, _clad_to_engine_cozmo, _clad_to_game_anki


#: Length of time in seconds to go without receiving an observed event before
#: assuming that Cozmo can no longer see an object.
OBJECT_VISIBILITY_TIMEOUT = 0.4


class EvtObjectObserved(event.Event):
    '''Triggered whenever an object is visually identified by the robot.

    A stream of these events are produced while an object is visible to the robot.
    Each event has an updated image_box field.

    See EvtObjectAppeared if you only want to know when an object first
    becomes visible.
    '''
    obj = 'The object that was observed'
    updated = 'A set of field names that have changed'
    image_box = 'A comzo.util.ImageBox defining where the object is within Cozmo\'s camera view'
    pose = 'The cozmo.util.Pose defining the position and rotation of the object'


class EvtObjectAppeared(event.Event):
    '''Triggered whenever an object is first visually identified by a robot.

    This differs from EvtObjectObserved in that it's only triggered when
    an object initially becomes visible.  If it disappears for more than
    OBJECT_VISIBILITY_TIMEOUT (0.2) seconds and then is seen again, a
    EvtObjectDisappeared will be dispatched, followed by another
    EvtObjectAppeared event.

    For continuous tracking information about a visible object, see
    EvtObjectObserved.
    '''
    obj = 'The object that was observed'
    updated = 'A set of field names that have changed'
    image_box = 'A comzo.util.ImageBox defining where the object is within Cozmo\'s camera view'
    pose = 'The cozmo.util.Pose defining the position and rotation of the object'


class EvtObjectAvailable(event.Event):
    '''Triggered when the engine reports that an object is available (i.e. exists).

    This will usually occur at the start of the program in response to the SDK
    sending RequestObjectStates to the engine.
    '''
    obj = 'The object that is available'
    updated = 'A set of field names that have changed'
    pose = 'The cozmo.util.Pose defining the position and rotation of the object'


class EvtObjectDisappeared(event.Event):
    '''Triggered whenever an object that was previously being observed is no longer visible.'''
    obj = 'The object that is no longer being observed'

class EvtObjectMoving(event.Event):
    'Triggered when an active object is currently moving.'
    obj = 'The object that is currently moving'
    # :class:`~cozmo.util.Vector3`: The currently measured acceleration
    acceleration = 'The currently measured acceleration'
    move_duration = 'The current duration of time (in seconds) that the object has spent moving'

class EvtObjectMovingStarted(event.Event):
    'Triggered when an active object starts moving.'
    obj = 'The object that started moving'
    #: :class:`~cozmo.util.Vector3`: The currently measured acceleration
    acceleration = 'The currently measured acceleration'

class EvtObjectMovingStopped(event.Event):
    'Triggered when an active object stops moving.'
    obj = 'The object that stopped moving'
    move_duration = 'The duration of time (in seconds) that the object spent moving'

class EvtObjectTapped(event.Event):
    'Triggered when an active object is tapped.'
    obj = 'The object that was tapped'
    tap_count = 'Number of taps detected'
    tap_duration = 'The duration of the tap in ms'
    tap_intensity = 'The intensity of the tap'


class EvtObjectConnectChanged(event.Event):
    'Triggered when an active object has connected or disconnected from the robot.'
    obj = 'The object that connected or disconnected'
    connected = 'True if the object connected, False if it disconnected'


class ObservableElement(event.Dispatcher):
    '''The base type for anything Cozmo can see.'''

    #: Length of time in seconds to go without receiving an observed event before
    #: assuming that Cozmo can no longer see an element. Can be overridden in sub
    #: classes.
    visibility_timeout = OBJECT_VISIBILITY_TIMEOUT

    def __init__(self, conn, world, robot, **kw):
        super().__init__(**kw)
        self._robot = robot
        self._pose = None
        self.conn = conn
        #: :class:`cozmo.world.World`: The robot's world in which this element is located.
        self.world = world

        #: float: The time the last event was received.
        #: ``None`` if no events have yet been received.
        self.last_event_time = None

        #: float: The time the element was last observed by the robot.
        #: ``None`` if the element has not yet been observed.
        self.last_observed_time = None

        #: int: The robot's timestamp of the last observed event.
        #: ``None`` if the element has not yet been observed.
        #: In milliseconds relative to robot epoch.
        self.last_observed_robot_timestamp = None

        #: :class:`~cozmo.util.ImageBox`: The ImageBox defining where the
        #: object was last visible within Cozmo's camera view.
        #: ``None`` if the element has not yet been observed.
        self.last_observed_image_box = None

        self._is_visible = False
        self._observed_timeout_handler = None

    def __repr__(self):
        extra = self._repr_values()
        if len(extra) > 0:
            extra = ' '+extra
        if self.pose:
            extra += ' pose=%s' % self.pose

        return '<%s%s is_visible=%s>' % (self.__class__.__name__,
                                         extra, self.is_visible)

    #### Private Methods ####

    def _repr_values(self):
        return ''

    def _update_field(self, changed, field_name, new_value):
        # Set only changed fields and update the passed in changed set
        current = getattr(self, field_name)
        if current != new_value:
            setattr(self, field_name, new_value)
            changed.add(field_name)

    def _reset_observed_timeout_handler(self):
        if self._observed_timeout_handler is not None:
            self._observed_timeout_handler.cancel()
        self._observed_timeout_handler = self._loop.call_later(
            self.visibility_timeout, self._observed_timeout)

    def _observed_timeout(self):
        # triggered when the element is no longer considered "visible"
        # ie. visibility_timeout seconds after the last observed event
        self._is_visible = False
        self._dispatch_disappeared_event()

    def _dispatch_observed_event(self, changed_fields, image_box):
        # Override in subclass if there is a specific event for that type
        pass

    def _dispatch_appeared_event(self, changed_fields, image_box):
        # Override in subclass if there is a specific event for that type
        pass

    def _dispatch_disappeared_event(self):
        # Override in subclass if there is a specific event for that type
        pass

    def _on_observed(self, image_box, timestamp, changed_fields):
        # Called from subclasses on their corresponding observed messages
        newly_visible = self._is_visible == False
        self._is_visible = True

        changed_fields |= {'last_observed_time', 'last_observed_robot_timestamp',
                           'last_event_time', 'last_observed_image_box'}

        now = time.time()
        self.last_observed_time = now
        self.last_observed_robot_timestamp = timestamp
        self.last_event_time = now
        self.last_observed_image_box = image_box
        self._reset_observed_timeout_handler()
        self._dispatch_observed_event(changed_fields, image_box)

        if newly_visible:
            self._dispatch_appeared_event(changed_fields, image_box)

    #### Properties ####

    @property
    def pose(self):
        ''':class:`cozmo.util.Pose`: The pose of the element in the world.

        Is ``None`` for elements that don't have pose information.
        '''
        return self._pose

    @property
    def time_since_last_seen(self):
        '''float: time since this element was last seen (math.inf if never)'''
        if self.last_observed_time is None:
            return math.inf
        return time.time() - self.last_observed_time

    @property
    def is_visible(self):
        '''bool: True if the element has been observed recently.

        "recently" is defined as :attr:`visibility_timeout` seconds.
        '''
        return self._is_visible


class ObservableObject(ObservableElement):
    '''The base type for objects in Cozmo's world.

    See parent class :class:`ObservableElement` for additional properties
    and methods.
    '''

    #: bool: True if this type of object can be physically picked up by Cozmo
    pickupable = False
    #: bool: True if this type of object can have objects physically placed on it by Cozmo
    place_objects_on_this = False

    def __init__(self, conn, world, object_id=None, **kw):
        super().__init__(conn, world, robot=None, **kw)
        self._object_id = object_id

    #### Private Methods ####

    def _repr_values(self):
        return 'object_id=%s' % self.object_id

    def _dispatch_observed_event(self, changed_fields, image_box):
        self.dispatch_event(EvtObjectObserved, obj=self,
                updated=changed_fields, image_box=image_box, pose=self._pose)

    def _dispatch_appeared_event(self, changed_fields, image_box):
        self.dispatch_event(EvtObjectAppeared, obj=self,
                updated=changed_fields, image_box=image_box, pose=self._pose)

    def _dispatch_disappeared_event(self):
        self.dispatch_event(EvtObjectDisappeared, obj=self)

    def _handle_object_state(self, object_state):
        # triggered when engine sends an ObjectStates message
        # as a response to a RequestObjectStates message
        if (self.last_observed_robot_timestamp and
                (self.last_observed_robot_timestamp > object_state.lastObservedTimestamp)):
            logger.debug("ignoring old object_state=%s obj=%s (last_observed_robot_timestamp=%s)",
                         object_state, self, self.last_observed_robot_timestamp)
            return

        changed_fields = {'last_observed_robot_timestamp', 'pose'}

        self.last_observed_robot_timestamp = object_state.lastObservedTimestamp

        self._pose = util.Pose._create_from_clad(object_state.pose)
        if object_state.poseState == _clad_to_game_anki.PoseState.Unknown:
            self._pose.invalidate()

        self.dispatch_event(EvtObjectAvailable,
                            obj=self,
                            updated=changed_fields,
                            pose=self._pose)

    #### Properties ####

    @property
    def object_id(self):
        '''int: The internal ID assigned to the object.

        This value can only be assigned once as it is static in the engine.
        '''
        return self._object_id

    @object_id.setter
    def object_id(self, value):
        if self._object_id is not None:
            # We cannot currently rely on Engine ensuring that object ID remains static
            # E.g. in the case of a cube disconnecting and reconnecting it's removed
            # and then re-added to blockworld which results in a new ID.
            logger.warning("Changing object_id for %s from %s to %s", self.__class__, self._object_id, value)
        else:
            logger.debug("Setting object_id for %s to %s", self.__class__, value)
        self._object_id = value

    #### Private Event Handlers ####

    def _recv_msg_object_connection_state(self, _, *, msg):
        if self.connected != msg.connected:
            self.connected = msg.connected
            self.dispatch_event(EvtObjectConnectChanged, obj=self, connected=self.connected)

    def _recv_msg_robot_observed_object(self, evt, *, msg):

        changed_fields = {'pose'}
        self._pose = util.Pose._create_from_clad(msg.pose)

        image_box = util.ImageBox._create_from_clad_rect(msg.img_rect)
        self._on_observed(image_box, msg.timestamp, changed_fields)

    #### Public Event Handlers ####

    #### Event Wrappers ####

    #### Commands ####


#: LightCube1Id's markers look a bit like a paperclip
LightCube1Id = _clad_to_game_cozmo.ObjectType.Block_LIGHTCUBE1
#: LightCube2Id's markers look a bit like a lamp
LightCube2Id = _clad_to_game_cozmo.ObjectType.Block_LIGHTCUBE2
#: LightCube3Id's markers look a bit like the letters 'ab' over 'T'
LightCube3Id = _clad_to_game_cozmo.ObjectType.Block_LIGHTCUBE3


class LightCube(ObservableObject):
    '''A light cube object has four LEDs that Cozmo can actively manipulate and communicate with.

    See parent class :class:`ObservableObject` for additional properties
    and methods.
    '''
    #TODO investigate why the top marker orientation of a cube is a bit strange

    #: Voltage where a cube's battery can be considered empty
    EMPTY_VOLTAGE = 1.0
    #: Voltage where a cube's battery can be considered full
    FULL_VOLTAGE = 1.5

    pickupable = True
    place_objects_on_this = True

    def __init__(self, *a, **kw):
        super().__init__(*a, **kw)

        #: float: The time the object was last tapped
        #: ``None`` if the cube wasn't tapped yet.
        self.last_tapped_time = None

        #: int: The robot's timestamp of the last tapped event.
        #: ``None`` if the cube wasn't tapped yet.
        #: In milliseconds relative to robot epoch.
        self.last_tapped_robot_timestamp = None

        #: float: The time the object was last moved
        #: ``None`` if the cube wasn't moved yet.
        self.last_moved_time = None

        #: float: The time the object started moving when last moved
        self.last_moved_start_time = None

        #: int: The robot's timestamp of the last move event.
        #: ``None`` if the cube wasn't moved yet.
        #: In milliseconds relative to robot epoch.
        self.last_moved_robot_timestamp = None

        #: int: The robot's timestamp of when the object started moving when last moved
        #: ``None`` if the cube wasn't moved yet.
        #: In milliseconds relative to robot epoch.
        self.last_moved_start_robot_timestamp = None

        #: float: Battery voltage.
        #: ``None`` if no voltage reading has been received yet
        self.battery_voltage = None

        #: bool: True if the cube's accelerometer indicates that the cube is moving.
        self.is_moving = False

    def _repr_values(self):
        super_values = super()._repr_values()
        if len(super_values) > 0:
            super_values += ' '
        return ('{super_values}'
                'battery={self.battery_str:s}'.format(self=self, super_values=super_values))

    #### Private Methods ####

    def _set_light(self, msg, idx, light):
        if not isinstance(light, lights.Light):
            raise TypeError("Expected a lights.Light")
        msg.onColor[idx] = light.on_color.int_color
        msg.offColor[idx] = light.off_color.int_color
        msg.onPeriod_ms[idx] = light.on_period_ms
        msg.offPeriod_ms[idx] = light.off_period_ms
        msg.transitionOnPeriod_ms[idx] = light.transition_on_period_ms
        msg.transitionOffPeriod_ms[idx] = light.transition_off_period_ms


    #### Event Wrappers ####

    async def wait_for_tap(self, timeout=None):
        '''Wait for the object to receive a tap event.

        Args:
            timeout (float): Maximum time to wait for a tap, in seconds.  None for indefinite
        Returns:
            A :class:`EvtObjectTapped` object if a tap was received.
        '''
        return await self.wait_for(EvtObjectTapped, timeout=timeout)


    #### Properties ####

    @property
    def battery_percentage(self):
        """float: Battery level as a percentage."""
        if self.battery_voltage is None:
            # not received a voltage measurement yet
            return None
        elif self.battery_voltage >= self.FULL_VOLTAGE:
            return 100.0
        elif self.battery_voltage <= self.EMPTY_VOLTAGE:
            return 0.0
        else:
            return 100.0 * ((self.battery_voltage - self.EMPTY_VOLTAGE) /
                            (self.FULL_VOLTAGE - self.EMPTY_VOLTAGE))

    @property
    def battery_str(self):
        """str: String representation of the battery level."""
        if self.battery_voltage is None:
            return "Unknown"
        else:
            return ('{self.battery_percentage:.0f}%'.format(self=self))

    #### Private Event Handlers ####
    def _recv_msg_object_tapped(self, evt, *, msg):
        now = time.time()
        self.last_event_time = now
        self.last_tapped_time = now
        self.last_tapped_robot_timestamp = msg.timestamp
        tap_intensity = msg.tapPos - msg.tapNeg
        self.dispatch_event(EvtObjectTapped, obj=self,
            tap_count=msg.numTaps, tap_duration=msg.tapTime, tap_intensity=tap_intensity)

    def _recv_msg_object_moved(self, evt, *, msg):
        now = time.time()
        started_moving = not self.is_moving
        self.is_moving = True
        self.last_event_time = now
        self.last_moved_time = now
        self.last_moved_robot_timestamp = msg.timestamp

        acceleration = util.Vector3(msg.accel.x, msg.accel.y, msg.accel.z)

        if started_moving:
            self.last_moved_start_time = now
            self.last_moved_start_robot_timestamp = msg.timestamp
            self.dispatch_event(EvtObjectMovingStarted, obj=self,
                                acceleration=acceleration)
        else:
            move_duration = now - self.last_moved_start_time
            self.dispatch_event(EvtObjectMoving, obj=self,
                                acceleration=acceleration,
                                move_duration=move_duration)

    def _recv_msg_object_stopped_moving(self, evt, *, msg):
        now = time.time()
        if self.is_moving:
            self.is_moving = False
            move_duration = now - self.last_moved_start_time
        else:
            # This happens for very short movements that are immediately
            # considered stopped (no acceleration info is present)
            move_duration = 0.0
        self.dispatch_event(EvtObjectMovingStopped, obj=self,
                            move_duration=move_duration)

    def _recv_msg_object_power_level(self, evt, *, msg):
        self.battery_voltage = msg.batteryLevel * 0.01
        pass

    #### Public Event Handlers ####

    def recv_evt_object_tapped(self, evt, **kw):
        pass

    #### Commands ####

    # TODO: make this explicit as to which light goes to which corner.
    def set_light_corners(self, light1, light2, light3, light4):
        """Set the light for each corner"""
        msg = _clad_to_engine_iface.SetAllActiveObjectLEDs(
                objectID=self.object_id, robotID=self._robot.robot_id)
        for i, light in enumerate( (light1, light2, light3, light4) ):
            if light is not None:
                lights._set_light(msg, i, light)

        self.conn.send_msg(msg)

    def set_lights(self, light):
        '''Set all lights on the cube

        Args:
            light (:class:`cozmo.lights.Light`): The settings for the lights.
        '''
        msg = _clad_to_engine_iface.SetAllActiveObjectLEDs(
                objectID=self.object_id, robotID=self._robot.robot_id)
        for i in range(4):
            lights._set_light(msg, i, light)

        self.conn.send_msg(msg)

    def set_lights_off(self):
        '''Turn off all the lights on the cube.'''
        self.set_lights(lights.off_light)


class Charger(ObservableObject):
    '''Cozmo's charger object, which the robot can observe and drive toward.

    See parent class :class:`ObservableObject` for additional properties
    and methods.
    '''

    def __init__(self, *a, **kw):
        super().__init__(*a, **kw)


class CustomObject(ObservableObject):
    '''An object defined by the SDK. It is bound to a specific objectType e.g ``CustomType00``.

    This defined object is given a size in the x,y and z axis. The dimensions
    of the markers on the object are also defined. We get an
    :class:`cozmo.objects.EvtObjectObserved` message when the robot sees these
    markers.

    See parent class :class:`ObservableObject` for additional properties
    and methods.
    '''

    def __init__(self, conn, world, object_type,
                 x_size_mm, y_size_mm, z_size_mm,
                 marker_width_mm, marker_height_mm, is_unique, **kw):
        super().__init__(conn, world, **kw)

        self.object_type = object_type
        self._x_size_mm = x_size_mm
        self._y_size_mm = y_size_mm
        self._z_size_mm = z_size_mm
        self._marker_width_mm = marker_width_mm
        self._marker_height_mm = marker_height_mm
        self._is_unique = is_unique

    def _repr_values(self):
        return ('object_type={self.object_type} '
                'x_size_mm={self.x_size_mm:.1f} '
                'y_size_mm={self.y_size_mm:.1f} '
                'z_size_mm={self.z_size_mm:.1f} '
                'is_unique={self.is_unique}'.format(self=self))

    #### Private Methods ####

    #### Event Wrappers ####
    #### Properties ####
    @property
    def x_size_mm(self):
        '''float: Size of this object in its X axis, in millimeters.'''
        return self._x_size_mm

    @property
    def y_size_mm(self):
        '''float: Size of this object in its Y axis, in millimeters.'''
        return self._y_size_mm

    @property
    def z_size_mm(self):
        '''float: Size of this object in its Z axis, in millimeters.'''
        return self._z_size_mm

    @property
    def marker_width_mm(self):
        '''float: Width in millimeters of the marker on this object.'''
        return self._marker_width_mm

    @property
    def marker_height_mm(self):
        '''float: Height in millimeters of the marker on this object.'''
        return self._marker_height_mm

    @property
    def is_unique(self):
        '''bool: True if there should only be one of this object type in the world.'''
        return self._is_unique

    #### Private Event Handlers ####

    #### Public Event Handlers ####

    #### Commands ####


class _CustomObjectType(collections.namedtuple('_CustomObjectType', 'name id')):
    # Tuple mapping between CLAD ActionResult name and ID
    # All instances will be members of ActionResults

    # Keep _ActionResult as lightweight as a normal namedtuple
    __slots__ = ()

    def __str__(self):
        return 'CustomObjectTypes.%s' % self.name


class CustomObjectTypes:
    '''Defines all available custom object types.

    For use with world.define_custom methods such as
    :meth:`cozmo.world.World.define_custom_box`,
    :meth:`cozmo.world.World.define_custom_cube`, and
    :meth:`cozmo.world.World.define_custom_wall`
    '''

    #: CustomType00 - the first custom object type
    CustomType00 = _CustomObjectType("CustomType00", _clad_to_engine_cozmo.ObjectType.CustomType00)

    #:
    CustomType01 = _CustomObjectType("CustomType01", _clad_to_engine_cozmo.ObjectType.CustomType01)

    #:
    CustomType02 = _CustomObjectType("CustomType02", _clad_to_engine_cozmo.ObjectType.CustomType02)

    #:
    CustomType03 = _CustomObjectType("CustomType03", _clad_to_engine_cozmo.ObjectType.CustomType03)

    #:
    CustomType04 = _CustomObjectType("CustomType04", _clad_to_engine_cozmo.ObjectType.CustomType04)

    #:
    CustomType05 = _CustomObjectType("CustomType05", _clad_to_engine_cozmo.ObjectType.CustomType05)

    #:
    CustomType06 = _CustomObjectType("CustomType06", _clad_to_engine_cozmo.ObjectType.CustomType06)

    #:
    CustomType07 = _CustomObjectType("CustomType07", _clad_to_engine_cozmo.ObjectType.CustomType07)

    #:
    CustomType08 = _CustomObjectType("CustomType08", _clad_to_engine_cozmo.ObjectType.CustomType08)

    #:
    CustomType09 = _CustomObjectType("CustomType09", _clad_to_engine_cozmo.ObjectType.CustomType09)

    #:
    CustomType10 = _CustomObjectType("CustomType10", _clad_to_engine_cozmo.ObjectType.CustomType10)

    #:
    CustomType11 = _CustomObjectType("CustomType11", _clad_to_engine_cozmo.ObjectType.CustomType11)

    #:
    CustomType12 = _CustomObjectType("CustomType12", _clad_to_engine_cozmo.ObjectType.CustomType12)

    #:
    CustomType13 = _CustomObjectType("CustomType13", _clad_to_engine_cozmo.ObjectType.CustomType13)

    #:
    CustomType14 = _CustomObjectType("CustomType14", _clad_to_engine_cozmo.ObjectType.CustomType14)

    #:
    CustomType15 = _CustomObjectType("CustomType15", _clad_to_engine_cozmo.ObjectType.CustomType15)

    #:
    CustomType16 = _CustomObjectType("CustomType16", _clad_to_engine_cozmo.ObjectType.CustomType16)

    #:
    CustomType17 = _CustomObjectType("CustomType17", _clad_to_engine_cozmo.ObjectType.CustomType17)

    #:
    CustomType18 = _CustomObjectType("CustomType18", _clad_to_engine_cozmo.ObjectType.CustomType18)

    #: CustomType19 - the last custom object type
    CustomType19 = _CustomObjectType("CustomType19", _clad_to_engine_cozmo.ObjectType.CustomType19)


_CustomObjectMarker = collections.namedtuple('_CustomObjectMarker', 'name id')

class CustomObjectMarkers:
    '''Defines all available custom object markers.

    For use with world.define_custom methods such as
    :meth:`cozmo.world.World.define_custom_box`,
    :meth:`cozmo.world.World.define_custom_cube`, and
    :meth:`cozmo.world.World.define_custom_wall`
    '''

    #: .. image:: ../images/custom_markers/SDK_2Circles.png
    Circles2 = _CustomObjectMarker("Circles2", _clad_to_engine_cozmo.CustomObjectMarker.Circles2)

    #: .. image:: ../images/custom_markers/SDK_3Circles.png
    Circles3 = _CustomObjectMarker("Circles3", _clad_to_engine_cozmo.CustomObjectMarker.Circles3)

    #: .. image:: ../images/custom_markers/SDK_4Circles.png
    Circles4 = _CustomObjectMarker("Circles4", _clad_to_engine_cozmo.CustomObjectMarker.Circles4)

    #: .. image:: ../images/custom_markers/SDK_5Circles.png
    Circles5 = _CustomObjectMarker("Circles5", _clad_to_engine_cozmo.CustomObjectMarker.Circles5)

    #: .. image:: ../images/custom_markers/SDK_2Diamonds.png
    Diamonds2 = _CustomObjectMarker("Diamonds2", _clad_to_engine_cozmo.CustomObjectMarker.Diamonds2)

    #: .. image:: ../images/custom_markers/SDK_3Diamonds.png
    Diamonds3 = _CustomObjectMarker("Diamonds3", _clad_to_engine_cozmo.CustomObjectMarker.Diamonds3)

    #: .. image:: ../images/custom_markers/SDK_4Diamonds.png
    Diamonds4 = _CustomObjectMarker("Diamonds4", _clad_to_engine_cozmo.CustomObjectMarker.Diamonds4)

    #: .. image:: ../images/custom_markers/SDK_5Diamonds.png
    Diamonds5 = _CustomObjectMarker("Diamonds5", _clad_to_engine_cozmo.CustomObjectMarker.Diamonds5)

    #: .. image:: ../images/custom_markers/SDK_2Hexagons.png
    Hexagons2 = _CustomObjectMarker("Hexagons2", _clad_to_engine_cozmo.CustomObjectMarker.Hexagons2)

    #: .. image:: ../images/custom_markers/SDK_3Hexagons.png
    Hexagons3 = _CustomObjectMarker("Hexagons3", _clad_to_engine_cozmo.CustomObjectMarker.Hexagons3)

    #: .. image:: ../images/custom_markers/SDK_4Hexagons.png
    Hexagons4 = _CustomObjectMarker("Hexagons4", _clad_to_engine_cozmo.CustomObjectMarker.Hexagons4)

    #: .. image:: ../images/custom_markers/SDK_5Hexagons.png
    Hexagons5 = _CustomObjectMarker("Hexagons5", _clad_to_engine_cozmo.CustomObjectMarker.Hexagons5)

    #: .. image:: ../images/custom_markers/SDK_2Triangles.png
    Triangles2 = _CustomObjectMarker("Triangles2", _clad_to_engine_cozmo.CustomObjectMarker.Triangles2)

    #: .. image:: ../images/custom_markers/SDK_3Triangles.png
    Triangles3 = _CustomObjectMarker("Triangles3", _clad_to_engine_cozmo.CustomObjectMarker.Triangles3)

    #: .. image:: ../images/custom_markers/SDK_4Triangles.png
    Triangles4 = _CustomObjectMarker("Triangles4", _clad_to_engine_cozmo.CustomObjectMarker.Triangles4)

    #: .. image:: ../images/custom_markers/SDK_5Triangles.png
    Triangles5 = _CustomObjectMarker("Triangles5", _clad_to_engine_cozmo.CustomObjectMarker.Triangles5)


class FixedCustomObject():
    '''A fixed object defined by the SDK. It is given a pose and x,y,z sizes.

    This object cannot be observed by the robot so its pose never changes.
    The position is static in Cozmo's world view; once instantiated, these
    objects never move. This could be used to make Cozmo aware of objects and
    know to plot a path around them even when they don't have any markers.
    '''

    is_visible = False

    def __init__(self, pose, x_size_mm, y_size_mm, z_size_mm, object_id, *a, **kw):
        super().__init__(*a, **kw)
        self._pose = pose
        self._object_id = object_id
        self._x_size_mm = x_size_mm
        self._y_size_mm = y_size_mm
        self._z_size_mm = z_size_mm

    def __repr__(self):
        return ('<%s pose=%s object_id=%d x_size_mm=%.1f y_size_mm=%.1f z_size_mm=%.1f=>' %
                                        (self.__class__.__name__, self.pose, self.object_id,
                                         self.x_size_mm, self.y_size_mm, self.z_size_mm))

    #### Private Methods ####
    #### Event Wrappers ####
    #### Properties ####
    @property
    def object_id(self):
        '''int: The internal ID assigned to the object.

        This value can only be assigned once as it is static in the engine.
        '''
        return self._object_id

    @object_id.setter
    def object_id(self, value):
        if self._object_id is not None:
            raise ValueError("Cannot change object ID once set (from %s to %s)" % (self._object_id, value))
        logger.debug("Updated object_id for %s from %s to %s", self.__class__, self._object_id, value)
        self._object_id = value

    @property
    def pose(self):
        ''':class:`cozmo.util.Pose`: The pose of the object in the world.'''
        return self._pose

    @property
    def x_size_mm(self):
        '''float: The length of the object in its X axis, in millimeters.'''
        return self._x_size_mm

    @property
    def y_size_mm(self):
        '''float: The length of the object in its Y axis, in millimeters.'''
        return self._y_size_mm

    @property
    def z_size_mm(self):
        '''float: The length of the object in its Z axis, in millimeters.'''
        return self._z_size_mm


    #### Private Event Handlers ####
    #### Public Event Handlers ####
    #### Commands ####



# Copyright (c) 2016 Anki, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License in the file LICENSE.txt or at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

''' Cozmo's OLED screen that displays his face - related functions and values.
'''

# __all__ should order by constants, event classes, other classes, functions.
__all__ = ['dimensions', 'convert_pixels_to_screen_data',
           'convert_image_to_screen_data']


SCREEN_WIDTH = 128
SCREEN_HALF_HEIGHT = 32
SCREEN_HEIGHT = SCREEN_HALF_HEIGHT * 2


def dimensions():
    '''Return the dimension (width, height) of the oled screen.

    Note: The screen is displayed interlaced, with only every other line displayed
    This alternates every time the image is changed (no longer than 30 seconds)
    to prevent screen burn-in. Therefore to ensure the image looks correct on
    either scan-line offset we use half the vertical resolution

    Returns:
        A tuple of ints (width, height)
    '''
    return SCREEN_WIDTH, SCREEN_HALF_HEIGHT


def convert_pixels_to_screen_data(pixel_data, image_width, image_height):
    '''Convert a sequence of pixel data to the correct format to display on Cozmo's face.

    Args:
        pixel_data (:class:`bytes`): sequence of pixel values, should be in binary (1s or 0s)
        image_width (int): width of the image defined by the pixel_data
        image_height (int): height of the image defined by the pixel_data

    Returns:
        A :class:`bytearray` object representing all of the pixels (8 pixels packed per byte)

    Raises:
        ValueError: Invalid Dimensions
        ValueError: Bad image_width
        ValueError: Bad image_height
    '''

    if len(pixel_data) != (image_width * image_height):
        raise ValueError('Invalid Dimensions: len(pixel_data) {0} != image_width={1} * image_height={2} (== {3})'.
            format(len(pixel_data), image_width, image_height, image_width * image_height))

    num_columns_per_pixel = int(SCREEN_WIDTH / image_width)
    num_rows_per_pixel = int(SCREEN_HEIGHT / image_height)

    if (image_width * num_columns_per_pixel) != SCREEN_WIDTH:
        raise ValueError('Bad image_width: image_width {0} must be an exact integer divisor of {1}'.
                         format(image_width, SCREEN_WIDTH))

    if (image_height * num_rows_per_pixel) != SCREEN_HEIGHT:
        raise ValueError('Bad image_height: image_height {0} must be an exact integer divisor of {1}'.
                         format(image_height, SCREEN_HEIGHT))

    pixel_chunks = zip(*[iter(pixel_data)] * 8)  # convert into 8 pixel chunks - we'll pack each as 1 byte later
    pixel_chunks_per_row = int(SCREEN_WIDTH / 8)  # 8 pixels per byte (pixel-chunk)

    result_bytes = bytearray()

    x = 0
    y = 0
    for pixel_chunk in pixel_chunks:
        # convert the 8 pixels in the chunk into bits to write out
        # write each pixel bit num_columns_per_pixel times in a row
        pixel_byte = 0
        for pixel in pixel_chunk:
            for _ in range(num_columns_per_pixel):
                pixel_byte <<= 1
                pixel_byte += pixel
                x += 1
                if (x % 8) == 0:
                    result_bytes.append(pixel_byte)
                    pixel_byte = 0

        # check if this is the end of a row
        if x == SCREEN_WIDTH:
            x = 0
            y += 1

        if (x == 0) and (num_rows_per_pixel > 0):
            # at the end of a row - copy that row for every extra row-per-pixel
            for _ in range(num_rows_per_pixel-1):
                start_of_last_row = len(result_bytes) - pixel_chunks_per_row
                result_bytes.extend(result_bytes[start_of_last_row:])

    return result_bytes


def convert_image_to_screen_data(image, invert_image=False, pixel_threshold=127):
    ''' Convert an image into the correct format to display on Cozmo's face.

    Args:
        image (:class:`~PIL.Image.Image`): The image to display on Cozmo's face
        invert_image (bool): If true then pixels darker than the threshold are set on
        pixel_threshold (int): The grayscale threshold for what to consider on or off (0..255)

    Returns:
        A :class:`bytearray` object representing all of the pixels (8 pixels packed per byte)
    '''

    # convert to grayscale
    grayscale_image = image.convert('L')

    # convert to binary white/black (1/0)
    if invert_image:
        def pixel_func(x): return 1 if x <= pixel_threshold else 0
    else:
        def pixel_func(x): return 1 if x >= pixel_threshold else 0
    bw = grayscale_image.point(pixel_func, '1')

    # convert to a flattened 1D bytes object of pixel values (1s or 0s in this case)
    pixel_data = bytes(bw.getdata())

    return convert_pixels_to_screen_data(pixel_data, image.width, image.height)
